# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for store commands."""
import argparse
import sys
from unittest import mock

import pytest

from charmcraft import env
from charmcraft.application.commands import FetchLibCommand
from charmcraft.cmdbase import JSON_FORMAT
from charmcraft.store.models import Library
from tests import factory


@pytest.fixture
def store_mock():
    """The fixture to fake the store layer in all the tests."""
    store_mock = mock.MagicMock()

    def validate_params(config, ephemeral=False, needs_auth=True):
        """Check that the store received the Charmhub configuration and ephemeral flag."""
        assert config == env.CharmhubConfig()
        assert isinstance(ephemeral, bool)
        assert isinstance(needs_auth, bool)
        return store_mock

    with mock.patch("charmcraft.application.commands.store.Store", validate_params):
        yield store_mock


# region fetch-lib tests
@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_simple_downloaded(emitter, store_mock, tmp_path, monkeypatch, config, formatted):
    """Happy path fetching the lib for the first time (downloading it)."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    lib_content = "some test content with uñicode ;)"
    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=7,
            lib_name="testlib",
            charm_name="testcharm",
        ),
    }
    store_mock.get_library.return_value = Library(
        lib_id=lib_id,
        content=lib_content,
        content_hash="abc",
        api=0,
        patch=7,
        lib_name="testlib",
        charm_name="testcharm",
    )

    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips(
            [{"charm_name": "testcharm", "lib_name": "testlib", "api": 0}]
        ),
        mock.call.get_library("testcharm", lib_id, 0),
    ]
    if formatted:
        expected = [
            {
                "charm_name": "testcharm",
                "library_name": "testlib",
                "library_id": lib_id,
                "api": 0,
                "fetched": {
                    "patch": 7,
                    "content_hash": "abc",
                },
            },
        ]
        emitter.assert_json_output(expected)
    else:
        expected = "Library charms.testcharm.v0.testlib version 0.7 downloaded."
        emitter.assert_message(expected)
    saved_file = tmp_path / "lib" / "charms" / "testcharm" / "v0" / "testlib.py"
    assert saved_file.read_text() == lib_content


def test_fetchlib_simple_dash_in_name(emitter, store_mock, tmp_path, monkeypatch, config):
    """Happy path fetching the lib for the first time (downloading it)."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    lib_content = "some test content with uñicode ;)"
    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=7,
            lib_name="testlib",
            charm_name="test-charm",
        ),
    }
    store_mock.get_library.return_value = Library(
        lib_id=lib_id,
        content=lib_content,
        content_hash="abc",
        api=0,
        patch=7,
        lib_name="testlib",
        charm_name="test-charm",
    )

    args = argparse.Namespace(library="charms.test_charm.v0.testlib", format=None)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips(
            [{"charm_name": "test-charm", "lib_name": "testlib", "api": 0}]
        ),
        mock.call.get_library("test-charm", lib_id, 0),
    ]
    expected = "Library charms.test_charm.v0.testlib version 0.7 downloaded."
    emitter.assert_message(expected)
    saved_file = tmp_path / "lib" / "charms" / "test_charm" / "v0" / "testlib.py"
    assert saved_file.read_text() == lib_content


def test_fetchlib_simple_dash_in_name_on_disk(emitter, store_mock, tmp_path, monkeypatch, config):
    """Happy path fetching the lib for the first time (downloading it)."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    lib_content = "test-content"
    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=7,
            lib_name="testlib",
            charm_name="test-charm",
        ),
    }
    store_mock.get_library.return_value = Library(
        lib_id=lib_id,
        content=lib_content,
        content_hash="abc",
        api=0,
        patch=7,
        lib_name="testlib",
        charm_name="test-charm",
    )
    factory.create_lib_filepath("test-charm", "testlib", api=0, patch=1, lib_id=lib_id)

    args = argparse.Namespace(library=None, format=None)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips([{"lib_id": "test-example-lib-id", "api": 0}]),
        mock.call.get_library("test-charm", lib_id, 0),
    ]
    expected = "Library charms.test_charm.v0.testlib updated to version 0.7."
    emitter.assert_message(expected)


def test_fetchlib_simple_updated(emitter, store_mock, tmp_path, monkeypatch, config):
    """Happy path fetching the lib for Nth time (updating it)."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    content, content_hash = factory.create_lib_filepath(
        "testcharm", "testlib", api=0, patch=1, lib_id=lib_id
    )

    new_lib_content = "some test content with uñicode ;)"
    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=2,
            lib_name="testlib",
            charm_name="testcharm",
        ),
    }
    store_mock.get_library.return_value = Library(
        lib_id=lib_id,
        content=new_lib_content,
        content_hash="abc",
        api=0,
        patch=2,
        lib_name="testlib",
        charm_name="testcharm",
    )

    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=None)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips([{"lib_id": lib_id, "api": 0}]),
        mock.call.get_library("testcharm", lib_id, 0),
    ]
    expected = "Library charms.testcharm.v0.testlib updated to version 0.2."
    emitter.assert_message(expected)
    saved_file = tmp_path / "lib" / "charms" / "testcharm" / "v0" / "testlib.py"
    assert saved_file.read_text() == new_lib_content


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_all(emitter, store_mock, tmp_path, monkeypatch, config, formatted):
    """Update all the libraries found in disk."""
    monkeypatch.chdir(tmp_path)

    c1, h1 = factory.create_lib_filepath(
        "testcharm1", "testlib1", api=0, patch=1, lib_id="lib_id_1"
    )
    c2, h2 = factory.create_lib_filepath(
        "testcharm2", "testlib2", api=3, patch=5, lib_id="lib_id_2"
    )

    store_mock.get_libraries_tips.return_value = {
        ("lib_id_1", 0): Library(
            lib_id="lib_id_1",
            content=None,
            content_hash="abc",
            api=0,
            patch=2,
            lib_name="testlib1",
            charm_name="testcharm1",
        ),
        ("lib_id_2", 3): Library(
            lib_id="lib_id_2",
            content=None,
            content_hash="def",
            api=3,
            patch=14,
            lib_name="testlib2",
            charm_name="testcharm2",
        ),
    }
    _store_libs_info = [
        Library(
            lib_id="lib_id_1",
            content="new lib content 1",
            content_hash="xxx",
            api=0,
            patch=2,
            lib_name="testlib1",
            charm_name="testcharm1",
        ),
        Library(
            lib_id="lib_id_2",
            content="new lib content 2",
            content_hash="yyy",
            api=3,
            patch=14,
            lib_name="testlib2",
            charm_name="testcharm2",
        ),
    ]
    store_mock.get_library.side_effect = lambda *a: _store_libs_info.pop(0)

    args = argparse.Namespace(library=None, format=formatted)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips(
            [
                {"lib_id": "lib_id_1", "api": 0},
                {"lib_id": "lib_id_2", "api": 3},
            ]
        ),
        mock.call.get_library("testcharm1", "lib_id_1", 0),
        mock.call.get_library("testcharm2", "lib_id_2", 3),
    ]
    names = [
        "charms.testcharm1.v0.testlib1",
        "charms.testcharm2.v3.testlib2",
    ]
    emitter.assert_debug("Libraries found under 'lib/charms': " + str(names))
    if formatted:
        expected = [
            {
                "charm_name": "testcharm1",
                "library_name": "testlib1",
                "library_id": "lib_id_1",
                "api": 0,
                "fetched": {
                    "patch": 2,
                    "content_hash": "xxx",
                },
            },
            {
                "charm_name": "testcharm2",
                "library_name": "testlib2",
                "library_id": "lib_id_2",
                "api": 3,
                "fetched": {
                    "patch": 14,
                    "content_hash": "yyy",
                },
            },
        ]
        emitter.assert_json_output(expected)
    else:
        emitter.assert_messages(
            [
                "Library charms.testcharm1.v0.testlib1 updated to version 0.2.",
                "Library charms.testcharm2.v3.testlib2 updated to version 3.14.",
            ]
        )

    saved_file = tmp_path / "lib" / "charms" / "testcharm1" / "v0" / "testlib1.py"
    assert saved_file.read_text() == "new lib content 1"
    saved_file = tmp_path / "lib" / "charms" / "testcharm2" / "v3" / "testlib2.py"
    assert saved_file.read_text() == "new lib content 2"


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_store_not_found(emitter, store_mock, config, formatted):
    """The indicated library is not found in the store."""
    store_mock.get_libraries_tips.return_value = {}
    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)
    FetchLibCommand(config).run(args)

    store_mock.get_libraries_tips.assert_called_once_with(
        [{"charm_name": "testcharm", "lib_name": "testlib", "api": 0}]
    ),
    error_message = "Library charms.testcharm.v0.testlib not found in Charmhub."
    if formatted:
        expected = [
            {
                "charm_name": "testcharm",
                "library_name": "testlib",
                "library_id": None,
                "api": 0,
                "error_message": error_message,
            },
        ]
        emitter.assert_json_output(expected)
    else:
        emitter.assert_message(error_message)


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_store_is_old(emitter, store_mock, tmp_path, monkeypatch, config, formatted):
    """The store has an older version that what is found locally."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    factory.create_lib_filepath("testcharm", "testlib", api=0, patch=7, lib_id=lib_id)

    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=6,
            lib_name="testlib",
            charm_name="testcharm",
        ),
    }
    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)
    FetchLibCommand(config).run(args)

    store_mock.get_libraries_tips.assert_called_once_with([{"lib_id": lib_id, "api": 0}])
    error_message = "Library charms.testcharm.v0.testlib has local changes, cannot be updated."
    if formatted:
        expected = [
            {
                "charm_name": "testcharm",
                "library_name": "testlib",
                "library_id": lib_id,
                "api": 0,
                "error_message": error_message,
            },
        ]
        emitter.assert_json_output(expected)
    else:
        emitter.assert_message(error_message)


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_store_same_versions_same_hash(
    emitter, store_mock, tmp_path, monkeypatch, config, formatted
):
    """The store situation is the same than locally."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    _, c_hash = factory.create_lib_filepath("testcharm", "testlib", api=0, patch=7, lib_id=lib_id)

    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash=c_hash,
            api=0,
            patch=7,
            lib_name="testlib",
            charm_name="testcharm",
        ),
    }
    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)
    FetchLibCommand(config).run(args)

    store_mock.get_libraries_tips.assert_called_once_with([{"lib_id": lib_id, "api": 0}])
    error_message = "Library charms.testcharm.v0.testlib was already up to date in version 0.7."
    if formatted:
        expected = [
            {
                "charm_name": "testcharm",
                "library_name": "testlib",
                "library_id": lib_id,
                "api": 0,
                "error_message": error_message,
            },
        ]
        emitter.assert_json_output(expected)
    else:
        emitter.assert_message(error_message)


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_fetchlib_store_same_versions_different_hash(
    emitter, store_mock, tmp_path, monkeypatch, config, formatted
):
    """The store has the lib in the same version, but with different content."""
    monkeypatch.chdir(tmp_path)

    lib_id = "test-example-lib-id"
    factory.create_lib_filepath("testcharm", "testlib", api=0, patch=7, lib_id=lib_id)

    store_mock.get_libraries_tips.return_value = {
        (lib_id, 0): Library(
            lib_id=lib_id,
            content=None,
            content_hash="abc",
            api=0,
            patch=7,
            lib_name="testlib",
            charm_name="testcharm",
        ),
    }
    args = argparse.Namespace(library="charms.testcharm.v0.testlib", format=formatted)
    FetchLibCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.get_libraries_tips([{"lib_id": lib_id, "api": 0}]),
    ]
    error_message = "Library charms.testcharm.v0.testlib has local changes, cannot be updated."
    if formatted:
        expected = [
            {
                "charm_name": "testcharm",
                "library_name": "testlib",
                "library_id": lib_id,
                "api": 0,
                "error_message": error_message,
            },
        ]
        emitter.assert_json_output(expected)
    else:
        emitter.assert_message(error_message)


# endregion



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for skopeo wrapper."""

import pathlib
import platform
from unittest import mock

import pytest

from charmcraft.utils.skopeo import Skopeo

pytestmark = [
    pytest.mark.xfail(
        platform.system().lower() not in ("linux", "darwin"),
        reason="Don't necessarily have skopeo on non Linux/mac platforms.",
        strict=False,  # Allow them to pass anyway.
    ),
]

IMAGE_PATHS = [  # See: https://github.com/containers/skopeo/blob/main/docs/skopeo.1.md#image-names
    "containers-storage:my/local:image",
    "dir:/tmp/some-dir",
    "docker://my-image:latest",
    "docker-archive:/tmp/some-archive",
    "docker-archive:/tmp/some-archive:latest",
    "docker-daemon:sha256:f515493110d497051b4a5c4d977c2b1e7f38190def919ab22683e6785b9d5067",
    "docker-daemon:ubuntu:24.04",
    "oci:/tmp/some-dir:latest",
    "oci-archive:my-image.tar",
]


@pytest.mark.parametrize("path", ["/skopeo", "/bin/skopeo"])
def test_skopeo_path(fake_process, path):
    fake_process.register([path, "--version"])
    skopeo = Skopeo(skopeo_path=path)

    assert skopeo.get_global_command() == [path]


def test_find_skopeo_success(fake_process):
    path = "/fake/path/to/skopeo"
    fake_process.register([path, "--version"])
    with mock.patch("shutil.which", return_value=path) as mock_which:
        skopeo = Skopeo()

    assert skopeo.get_global_command() == [path]
    mock_which.assert_called_once_with("skopeo")


@pytest.mark.parametrize(
    ("kwargs", "expected"),
    [
        pytest.param({}, [], id="empty"),
        pytest.param({"insecure_policy": True}, ["--insecure-policy"], id="insecure_policy"),
        pytest.param({"arch": "amd64"}, ["--override-arch", "amd64"], id="amd64"),
        pytest.param({"arch": "arm64"}, ["--override-arch", "arm64"], id="arm64"),
        pytest.param({"arch": "riscv64"}, ["--override-arch", "riscv64"], id="riscv64"),
        pytest.param({"os": "linux"}, ["--override-os", "linux"], id="os-linux"),
        pytest.param({"os": "bsd"}, ["--override-os", "bsd"], id="os-bsd"),
        pytest.param(
            {"tmpdir": pathlib.Path("/tmp/skopeo_tmp")},
            ["--tmpdir", "/tmp/skopeo_tmp"],
            id="tmpdir",
        ),
    ],
)
def test_get_global_command(fake_process, kwargs, expected):
    """Tests for getting the global command and arguments."""
    fake_process.register(["/skopeo", "--version"])
    skopeo = Skopeo(skopeo_path="/skopeo", **kwargs)

    assert skopeo.get_global_command() == ["/skopeo", *expected]


@pytest.fixture
def fake_skopeo(fake_process):
    fake_process.register(["/skopeo", "--version"])
    return Skopeo(skopeo_path="/skopeo")


@pytest.mark.parametrize("source_image", IMAGE_PATHS)
@pytest.mark.parametrize("destination_image", IMAGE_PATHS)
@pytest.mark.parametrize(
    ("kwargs", "expected_args"),
    [
        ({}, []),
        ({"all_images": True}, ["--all"]),
        ({"preserve_digests": True}, ["--preserve-digests"]),
        ({"source_username": "user"}, ["--src-creds", "user"]),
        ({"source_password": "pass"}, ["--src-password", "pass"]),
        ({"source_username": "user", "source_password": "pass"}, ["--src-creds", "user:pass"]),
        ({"dest_username": "user"}, ["--dest-creds", "user"]),
        ({"dest_password": "pass"}, ["--dest-password", "pass"]),
        ({"dest_username": "user", "dest_password": "pass"}, ["--dest-creds", "user:pass"]),
    ],
)
def test_get_copy_command(
    fake_process, fake_skopeo: Skopeo, source_image, destination_image, kwargs, expected_args
):
    fake_process.register(
        [
            *fake_skopeo.get_global_command(),
            "copy",
            *expected_args,
            source_image,
            destination_image,
        ]
    )
    result = fake_skopeo.copy(source_image, destination_image, **kwargs)

    result.check_returncode()



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import argparse
from textwrap import dedent
from typing import Any

import pytest
from overrides import override

from charmcraft import extensions
from charmcraft.application.commands.extensions import ListExtensionsCommand
from tests.extensions.test_extensions import FakeExtension


class MyFakeExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "test-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {"terms": ["https://example.com/test"]}


class YourFakeExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "test-extension-2"
    bases = [("ubuntu", "23.04")]

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {"terms": ["https://example.com/test2"]}


@pytest.fixture
def fake_extensions(stub_extensions):
    extensions.register(MyFakeExtension.name, MyFakeExtension)
    extensions.register(YourFakeExtension.name, YourFakeExtension)


def test_expand_extensions_simple(fake_extensions, emitter):
    """List extensions"""
    cmd = ListExtensionsCommand(None)
    cmd.run(argparse.Namespace(format=None))
    emitter.assert_message(
        dedent(
            """\
            Extension name    Supported bases    Experimental bases
            ----------------  -----------------  --------------------
            test-extension    ubuntu@22.04
            test-extension-2  ubuntu@23.04"""
        )
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Internal models for store data structiues."""

import dataclasses
import datetime
import enum
import functools
from dataclasses import dataclass
from typing import Any, Literal, TypedDict

from craft_cli import CraftError
from typing_extensions import NotRequired, Self


@dataclasses.dataclass(frozen=True)
class Account:
    """Charmcraft-specific store account model.

    Deprecated in favour of implementation in craft-store.
    """

    name: str
    username: str
    id: str


@dataclasses.dataclass(frozen=True)
class Package:
    """Charmcraft-specific store package model.

    Deprecated in favour of implementation in craft-store.
    """

    id: str | None
    name: str
    type: Literal["charm", "bundle"]


@dataclasses.dataclass(frozen=True)
class MacaroonInfo:
    """Charmcraft-specific macaroon information model.

    Deprecated in favour of implementation in craft-store.
    """

    account: Account
    channels: list[str] | None
    packages: list[Package] | None
    permissions: list[str]


@dataclasses.dataclass(frozen=True)
class Entity:
    """Charmcraft-specific store entity model.

    Deprecated in favour of implementation in craft-store.
    """

    entity_type: Literal["charm", "bundle"]
    name: str
    private: bool
    status: str
    publisher_display_name: str


@dataclasses.dataclass(frozen=True)
class Error:
    """Charmcraft-specific store error model.

    Deprecated in favour of implementation in craft-store.
    """

    message: str
    code: str


@dataclasses.dataclass(frozen=True)
class Uploaded:
    """Charmcraft-specific store upload result model.

    Deprecated in favour of implementation in craft-store.
    """

    ok: bool
    status: int
    revision: int
    errors: list[Error]


@dataclasses.dataclass(frozen=True)
class Base:
    """Charmcraft-specific store object base model.

    Deprecated in favour of implementation in craft-store.
    """

    architecture: str
    channel: str
    name: str


@dataclasses.dataclass(frozen=True)
class Revision:
    """Charmcraft-specific store name revision model.

    Deprecated in favour of implementation in craft-store.
    """

    revision: int
    version: str | None
    created_at: datetime.datetime
    status: str
    errors: list[Error]
    bases: list[Base]


@dataclasses.dataclass(frozen=True)
class Resource:
    """Charmcraft-specific store name resource model.

    Deprecated in favour of implementation in craft-store.
    """

    name: str
    optional: bool
    revision: int
    resource_type: str


@dataclasses.dataclass(frozen=True)
class Release:
    """Charmcraft-specific store release model.

    Deprecated in favour of implementation in craft-store.
    """

    revision: int
    channel: str
    expires_at: datetime.datetime
    resources: list[Resource]
    base: Base


@dataclasses.dataclass(frozen=True)
class Channel:
    """Charmcraft-specific store channel model.

    Deprecated in favour of implementation in craft-store.
    """

    name: str
    fallback: str
    track: str
    risk: str
    branch: str


@dataclasses.dataclass(frozen=True)
class Library:
    """Charmcraft-specific store library model."""

    lib_id: str
    lib_name: str
    charm_name: str
    api: int
    patch: int
    content: str | None
    content_hash: str

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Convert a dictionary of this type to the type itself.

        Dictionary should match a single item of the `libraries` attribute in
        the response to fetch_libraries or fetch_library
        http://api.charmhub.io/docs/libraries.html#fetch_libraries
        """
        return cls(
            charm_name=value["charm-name"],
            lib_name=value["library-name"],
            lib_id=value["library-id"],
            api=value["api"],
            patch=value["patch"],
            content_hash=value["hash"],
            content=value.get("content"),
        )


@dataclasses.dataclass(frozen=True)
class RegistryCredentials:
    """Charmcraft-specific store registry credential model.

    Deprecated in favour of implementation in craft-store.
    """

    image_name: str
    username: str
    password: str


@functools.total_ordering
@enum.unique
class Risk(enum.Enum):
    """Standard risk tracks for a channel, orderable but not comparable to an int."""

    STABLE = 0
    CANDIDATE = 1
    BETA = 2
    EDGE = 3

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value
        return NotImplemented


@dataclass(frozen=True)
class ChannelData:
    """Data class for a craft store channel."""

    track: str | None
    risk: Risk
    branch: str | None

    @classmethod
    def from_str(cls, name: str):
        """Parse a channel name from a string using the standard store semantics.

        https://snapcraft.io/docs/channels
        """
        invalid_channel_error = CraftError(f"Invalid channel name: {name!r}")
        parts = name.split("/")
        track: str | None = None
        branch: str | None = None
        if len(parts) == 1:  # Just the risk, e.g. "stable"
            try:
                risk = Risk[parts[0].upper()]
            except KeyError:
                raise invalid_channel_error from None
        elif len(parts) == 2:
            try:  # risk/branch, e.g. "stable/debug"
                risk = Risk[parts[0].upper()]
                branch = parts[1]
            except KeyError:
                try:  # track/risk, e.g. "latest/stable"
                    risk = Risk[parts[1].upper()]
                    track = parts[0]
                except KeyError:
                    raise invalid_channel_error from None
        elif len(parts) == 3:  # Fully defined, e.g. "latest/stable/debug"
            try:
                risk = Risk[parts[1].upper()]
            except KeyError:
                raise invalid_channel_error from None
            track, _, branch = parts
        else:
            raise invalid_channel_error
        return cls(track, risk, branch)

    @property
    def name(self) -> str:
        """Get the channel name as a string."""
        risk = self.risk.name.lower()
        return "/".join(i for i in (self.track, risk, self.branch) if i is not None)


LibraryMetadataRequest = TypedDict(
    "LibraryMetadataRequest",
    {
        "charm-name": str,
        "library-name": str,
        "api": int,
        "patch": NotRequired[int],
    },
)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Service class for creating providers."""
from __future__ import annotations

import os
import pathlib

import craft_providers
from craft_application import services
from craft_providers import bases

from charmcraft import env


class ProviderService(services.ProviderService):
    """Business logic for getting providers."""

    def setup(self) -> None:
        """Set up the provider service for Charmcraft."""
        super().setup()

        # Forward all charmcraft environment variables
        for key, value in os.environ.items():
            if key.startswith("CHARMCRAFT_"):
                self.environment[key] = value

        self.environment["CHARMCRAFT_MANAGED_MODE"] = "1"

    def get_base(
        self,
        base_name: bases.BaseName,
        *,
        instance_name: str,
        **kwargs: bool | str | None | pathlib.Path,
    ) -> craft_providers.Base:
        """Get the base configuration from a base name.

        :param base_name: The base to lookup.
        :param instance_name: A name to assign to the instance.
        :param kwargs: Additional keyword arguments are sent directly to the base.

        If no cache_path is included, adds one.
        """
        # Forward the shared cache path.
        if "cache_path" not in kwargs:
            kwargs["cache_path"] = env.get_host_shared_cache_path()
        return super().get_base(
            base_name,
            instance_name=instance_name,
            # craft-application annotation is incorrect
            **kwargs,  # type: ignore[arg-type]
        )



# Copyright 2020-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Collection of utilities for charmcraft."""

from charmcraft.utils.charmlibs import (
    LibData,
    LibInternals,
    get_name_from_metadata,
    create_charm_name_from_importable,
    create_importable_name,
    get_lib_internals,
    get_lib_path,
    get_lib_module_name,
    get_lib_info,
    get_libs_from_tree,
    collect_charmlib_pydeps,
)
from charmcraft.utils.cli import (
    SingleOptionEnsurer,
    OutputFormat,
    ResourceOption,
    ChoicesList,
    confirm_with_user,
    format_content,
    format_timestamp,
    humanize_list,
)
from charmcraft.utils.platform import (
    OSPlatform,
    get_os_platform,
    validate_architectures,
)
from charmcraft.utils.file import S_IRALL, S_IXALL, make_executable, useful_filepath, build_zip
from charmcraft.utils.package import (
    get_pypi_packages,
    PACKAGE_LINE_REGEX,
    get_package_names,
    exclude_packages,
    get_pip_command,
    get_pip_version,
    get_requirements_file_package_names,
    validate_strict_dependencies,
)
from charmcraft.utils.project import (
    find_charm_sources,
    get_charm_name_from_path,
    get_templates_environment,
)
from charmcraft.utils.skopeo import Skopeo
from charmcraft.utils.store import get_packages
from charmcraft.utils.yaml import dump_yaml, load_yaml

__all__ = [
    "LibData",
    "LibInternals",
    "get_name_from_metadata",
    "create_charm_name_from_importable",
    "create_importable_name",
    "get_lib_internals",
    "get_lib_path",
    "get_lib_module_name",
    "get_lib_info",
    "get_libs_from_tree",
    "collect_charmlib_pydeps",
    "OSPlatform",
    "get_os_platform",
    "validate_architectures",
    "S_IRALL",
    "S_IXALL",
    "make_executable",
    "useful_filepath",
    "build_zip",
    "PACKAGE_LINE_REGEX",
    "format_timestamp",
    "get_pypi_packages",
    "get_package_names",
    "exclude_packages",
    "get_pip_command",
    "get_pip_version",
    "get_requirements_file_package_names",
    "validate_strict_dependencies",
    "SingleOptionEnsurer",
    "OutputFormat",
    "ResourceOption",
    "ChoicesList",
    "confirm_with_user",
    "format_content",
    "humanize_list",
    "find_charm_sources",
    "get_charm_name_from_path",
    "get_templates_environment",
    "Skopeo",
    "get_packages",
    "dump_yaml",
    "load_yaml",
]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft commands."""
import craft_application

from charmcraft.application.commands.analyse import Analyse, Analyze
from charmcraft.application.commands.extensions import (
    ListExtensionsCommand,
    ExtensionsCommand,
    ExpandExtensionsCommand,
)
from charmcraft.application.commands.init import InitCommand
from charmcraft.application.commands.lifecycle import (
    get_lifecycle_commands,
    PackCommand,
)
from charmcraft.application.commands.remote import RemoteBuild
from charmcraft.application.commands.store import (
    # auth
    FetchLibs,
    LoginCommand,
    LogoutCommand,
    WhoamiCommand,
    # name handling
    RegisterCharmNameCommand,
    RegisterBundleNameCommand,
    UnregisterNameCommand,
    ListNamesCommand,
    # pushing files and checking revisions
    UploadCommand,
    ListRevisionsCommand,
    # release process, and show status
    ReleaseCommand,
    PromoteBundleCommand,
    StatusCommand,
    CloseCommand,
    # libraries support
    CreateLibCommand,
    PublishLibCommand,
    ListLibCommand,
    FetchLibCommand,
    # resources support
    ListResourcesCommand,
    ListResourceRevisionsCommand,
    SetResourceArchitecturesCommand,
    UploadResourceCommand,
)
from charmcraft.application.commands.test import TestCommand
from charmcraft.application.commands.version import Version


def fill_command_groups(app: craft_application.Application) -> None:
    """Fill in all the command groups for Charmcraft."""
    app.add_command_group("Basic", [InitCommand])
    app.add_command_group("Lifecycle", [*get_lifecycle_commands(), RemoteBuild])
    app.add_command_group(
        "Store (account)",
        [
            # auth
            LoginCommand,
            LogoutCommand,
            WhoamiCommand,
            # name handling
            RegisterCharmNameCommand,
            RegisterBundleNameCommand,
            UnregisterNameCommand,
            ListNamesCommand,
        ],
    )
    app.add_command_group(
        "Store (charm or bundle)",
        [
            # pushing files and checking revisions
            UploadCommand,
            ListRevisionsCommand,
            # release process, and show status
            ReleaseCommand,
            PromoteBundleCommand,
            StatusCommand,
            CloseCommand,
            # resources support
            ListResourcesCommand,
            ListResourceRevisionsCommand,
            SetResourceArchitecturesCommand,
            UploadResourceCommand,
        ],
    )
    app.add_command_group(
        "Store (libraries)",
        [
            CreateLibCommand,
            PublishLibCommand,
            ListLibCommand,
            FetchLibs,
            FetchLibCommand,
        ],
    )
    app.add_command_group(
        "Extensions", [ExpandExtensionsCommand, ExtensionsCommand, ListExtensionsCommand]
    )
    app.add_command_group(
        "Other",
        [
            Analyse,
            Analyze,
            TestCommand,
            Version,
        ],
    )


__all__ = [
    "Analyse",
    "Analyze",
    "ListExtensionsCommand",
    "ExpandExtensionsCommand",
    "ExtensionsCommand",
    "InitCommand",
    "get_lifecycle_commands",
    "PackCommand",
    "LoginCommand",
    "LogoutCommand",
    "WhoamiCommand",
    "RegisterCharmNameCommand",
    "RegisterBundleNameCommand",
    "UnregisterNameCommand",
    "ListNamesCommand",
    "UploadCommand",
    "ListRevisionsCommand",
    "ReleaseCommand",
    "PromoteBundleCommand",
    "StatusCommand",
    "CloseCommand",
    "CreateLibCommand",
    "PublishLibCommand",
    "ListLibCommand",
    "FetchLibCommand",
    "ListResourcesCommand",
    "ListResourceRevisionsCommand",
    "SetResourceArchitecturesCommand",
    "TestCommand",
    "UploadResourceCommand",
]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""craft-application based lifecycle commands."""
from __future__ import annotations

import pathlib
import sys
import textwrap
from typing import TYPE_CHECKING, Any, cast

import craft_cli
from craft_application.commands import lifecycle
from craft_cli import ArgumentParsingError, CraftError
from typing_extensions import override

from charmcraft import models, services, utils

if TYPE_CHECKING:  # pragma: no cover
    import argparse

BUNDLE_MANDATORY_FILES = ["bundle.yaml", "README.md"]


def get_lifecycle_commands() -> list[type[craft_cli.BaseCommand]]:
    """Return the lifecycle related command group."""
    return [
        lifecycle.CleanCommand,
        lifecycle.PullCommand,
        lifecycle.BuildCommand,
        lifecycle.StageCommand,
        lifecycle.PrimeCommand,
        PackCommand,
    ]


class PackCommand(lifecycle.PackCommand):
    """Command to pack the final artifact."""

    name = "pack"
    help_msg = "Build the charm or bundle"
    overview = textwrap.dedent(
        """
        Build and pack a charm operator package or a bundle.

        You can `juju deploy` the resulting `.charm` or bundle's `.zip`
        file directly, or upload it to Charmhub with `charmcraft upload`.

        For the charm you must be inside a charm directory with a valid
        `metadata.yaml`, `requirements.txt` including the `ops` package
        for the Python operator framework, and an operator entrypoint,
        usually `src/charm.py`.  See `charmcraft init` to create a
        template charm directory structure.

        For the bundle you must already have a `bundle.yaml` (can be
        generated by Juju) and a README.md file.
        """
    )

    @override
    def _fill_parser(self, parser: argparse.ArgumentParser) -> None:
        super()._fill_parser(parser)

        parser.add_argument(
            "--bases-index",
            action="append",
            type=int,
            help="Index of 'bases' configuration to build (can be used multiple "
            "times); zero-based, defaults to all",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force packing even after finding lint errors",
        )
        parser.add_argument(
            "--measure",
            type=pathlib.Path,
            help="Dump measurements to the specified file",
        )
        include_charm_group = parser.add_mutually_exclusive_group()
        include_charm_group.add_argument(
            "--include-all-charms",
            action="store_true",
            help="For bundles, pack all charms whose source is inside the bundle directory",
        )
        include_charm_group.add_argument(
            "--include-charm",
            action="append",
            type=pathlib.Path,
            help="For bundles, pack the charm in the referenced path. Can be used multiple times",
        )
        parser.add_argument(
            "--output-bundle",
            type=pathlib.Path,
            help="Write the bundle configuration to this path",
        )
        parser.add_argument(
            "--format",
            choices=["json"],
            help="Produce a machine-readable format (currently only json)",
        )
        parser.add_argument(
            "--project-dir",
            "-p",
            type=pathlib.Path,
            default=pathlib.Path.cwd(),
            help="Specify the project's directory (defaults to current)",
        )

    @override
    @staticmethod
    def _should_add_shell_args() -> bool:
        return True

    def _validate_args(self, parsed_args: argparse.Namespace) -> None:
        project = cast(models.CharmcraftProject, self._services.project)
        package_service = cast(services.PackageService, self._services.package)
        if project.type == "charm":
            if parsed_args.include_all_charms:
                raise ArgumentParsingError(
                    "--include-all-charms can only be used when packing a bundle. "
                    f"Currently trying to pack: {package_service.project_dir}"
                )
            if parsed_args.include_charm:
                raise ArgumentParsingError(
                    "--include-charm can only be used when packing a bundle. "
                    f"Currently trying to pack: {package_service.project_dir}"
                )
            if parsed_args.output_bundle:
                raise ArgumentParsingError(
                    "--output-bundle can only be used when packing a bundle. "
                    f"Currently trying to pack: {package_service.project_dir}"
                )

    def _validate_bases_indices(self, bases_indices):
        """Validate that bases index is valid."""
        if bases_indices is None:
            return

        project = cast(models.Charm, self._services.project)

        msg = "Bases index '{}' is invalid (must be >= 0 and fit in configured bases)."
        len_configured_bases = len(project.bases)
        for bases_index in bases_indices:
            if bases_index < 0:
                raise CraftError(msg.format(bases_index))
            if bases_index >= len_configured_bases:
                raise CraftError(msg.format(bases_index))

    def run_managed(self, parsed_args: argparse.Namespace) -> bool:
        """Whether to run this command in managed mode.

        If we're packing a bundle, run unmanaged. Otherwise, do what other lifecycle
        commands do.
        """
        project_dir = pathlib.Path(getattr(parsed_args, "project_dir", "."))
        charmcraft_yaml = utils.load_yaml(project_dir / "charmcraft.yaml")
        # Always use a runner on non-Linux platforms.
        # Craft-parts is not designed to work on non-posix platforms, and most
        # notably here, the bundle plugin doesn't work on Windows.
        if sys.platform == "linux" and charmcraft_yaml and charmcraft_yaml.get("type") == "bundle":
            return False
        return super().run_managed(parsed_args)

    def _run(
        self,
        parsed_args: argparse.Namespace,
        step_name: str | None = None,
        **kwargs: Any,  # noqa: ANN401 (allow dynamic typing)
    ) -> None:
        self._validate_args(parsed_args)
        return super()._run(parsed_args, step_name, **kwargs)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for CharmBuilder."""


import pathlib
import sys

import pytest

from charmcraft import charm_builder

pytestmark = pytest.mark.skipif(
    sys.platform != "linux", reason="The charm builder only runs in managed mode."
)


@pytest.mark.parametrize(
    "requirements",
    [
        ["ops==2.15.0"],  # Requires pyyaml and websocket-client
    ],
)
def test_install_strict_dependencies_pip_check_error(
    monkeypatch, new_path: pathlib.Path, requirements: list[str]
):
    build_dir = new_path / "build"
    install_dir = new_path / "install"
    entrypoint = build_dir / "entrypoint.py"

    build_dir.mkdir()
    install_dir.mkdir()
    monkeypatch.chdir(build_dir)

    requirements_file = build_dir / "requirements.txt"
    requirements_file.write_text("\n".join(requirements))

    builder = charm_builder.CharmBuilder(
        builddir=build_dir,
        installdir=install_dir,
        entrypoint=entrypoint,
        requirements=[requirements_file],
        strict_dependencies=True,
    )

    with pytest.raises(RuntimeError, match="failed with retcode 1"):
        builder.handle_dependencies()


@pytest.mark.parametrize(
    "requirements",
    [
        ["distro==1.9.0"],  # No dependencies
    ],
)
def test_install_strict_dependencies_pip_check_success(
    monkeypatch, new_path: pathlib.Path, requirements: list[str]
):
    build_dir = new_path / "build"
    install_dir = new_path / "install"
    entrypoint = build_dir / "entrypoint.py"

    build_dir.mkdir()
    install_dir.mkdir()
    monkeypatch.chdir(build_dir)

    requirements_file = build_dir / "requirements.txt"
    requirements_file.write_text("\n".join(requirements))

    builder = charm_builder.CharmBuilder(
        builddir=build_dir,
        installdir=install_dir,
        entrypoint=entrypoint,
        requirements=[requirements_file],
        strict_dependencies=True,
    )

    builder.handle_dependencies()



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Configuration for services integration tests."""
import contextlib
import sys

import pyfakefs.fake_filesystem
import pytest

from charmcraft import services
from charmcraft.application.main import APP_METADATA, Charmcraft


@pytest.fixture
def service_factory(
    fs: pyfakefs.fake_filesystem.FakeFilesystem, fake_path, simple_charm
) -> services.CharmcraftServiceFactory:
    fake_project_dir = fake_path / "project"
    fake_project_dir.mkdir()

    # Allow access to the real venv library path.
    # This is necessary because certifi lazy-loads the certificate file.
    for python_path in sys.path:
        if not python_path:
            continue
        with contextlib.suppress(OSError):
            fs.add_real_directory(python_path)

    factory = services.CharmcraftServiceFactory(app=APP_METADATA)

    app = Charmcraft(app=APP_METADATA, services=factory)

    app._configure_services(provider_name=None)

    factory.project = simple_charm

    return factory



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft configuration pydantic model."""
from typing import TypedDict, cast

import pydantic
from craft_application import util
from craft_application.models import CraftBaseModel
from typing_extensions import Self

from charmcraft.models.basic import AttributeName, LinterName


class BaseDict(TypedDict, total=False):
    """TypedDict that describes only one base.

    This is equivalent to the short form base definition.
    """

    name: str
    channel: str
    architectures: list[str]


LongFormBasesDict = TypedDict(
    "LongFormBasesDict", {"build-on": list[BaseDict], "run-on": list[BaseDict]}
)


class Charmhub(CraftBaseModel):
    """Definition of Charmhub endpoint configuration."""

    api_url: pydantic.HttpUrl = cast(pydantic.HttpUrl, "https://api.charmhub.io")
    storage_url: pydantic.HttpUrl = cast(pydantic.HttpUrl, "https://storage.snapcraftcontent.com")
    registry_url: pydantic.HttpUrl = cast(pydantic.HttpUrl, "https://registry.jujucharms.com")


class Base(CraftBaseModel):
    """Represents a base."""

    name: pydantic.StrictStr
    channel: pydantic.StrictStr
    architectures: list[pydantic.StrictStr] = [util.get_host_architecture()]

    @classmethod
    def from_str_and_arch(cls, base_str: str, architectures: list[str]) -> Self:
        """Get a Base from a base string and list of architectures.

        :param base_str: A base string along the lines of "<name>@<channel>"
        :param architectures: A list of architectures (or ["all"])
        """
        name, _, channel = base_str.partition("@")
        return cls(name=name, channel=channel, architectures=architectures)


class BasesConfiguration(CraftBaseModel):
    """Definition of build-on/run-on combinations.

    Example::

        bases:
          - build-on:
              - name: ubuntu
                channel: "20.04"
                architectures: [amd64, arm64]
            run-on:
              - name: ubuntu
                channel: "20.04"
                architectures: [amd64, arm64]
              - name: ubuntu
                channel: "22.04"
                architectures: [amd64, arm64]
    """

    build_on: list[Base]
    run_on: list[Base]

    @pydantic.model_validator(mode="before")
    def _expand_base(cls, base: BaseDict | LongFormBasesDict) -> LongFormBasesDict:
        """Expand short-form bases into long-form bases."""
        if "build-on" in base:  # Assume long-form base already.
            return cast(LongFormBasesDict, base)
        return cast(LongFormBasesDict, {"build-on": [base], "run-on": [base]})


class Ignore(CraftBaseModel):
    """Definition of `analysis.ignore` configuration."""

    attributes: list[AttributeName] = []
    linters: list[LinterName] = []


class AnalysisConfig(CraftBaseModel):
    """Definition of `analysis` configuration."""

    ignore: Ignore = Ignore()


class Links(CraftBaseModel):
    """Definition of `links` in metadata."""

    contact: pydantic.StrictStr | list[pydantic.StrictStr] | None = None
    """Instructions for contacting the owner of the charm."""
    documentation: pydantic.AnyHttpUrl | None = None
    """The URL of the documentation for this charm."""
    issues: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None
    """A link to the issue tracker for this charm."""
    source: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None
    """Where to find this charm's source code."""
    website: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None
    """The website for this charm."""



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import itertools
import os
import pathlib
import re

import pytest


def get_python_filepaths(*, roots=None, python_paths=None):
    """Helper to retrieve paths of Python files."""
    if python_paths is None:
        python_paths = ["setup.py"]
    if roots is None:
        roots = ["charmcraft", "tests"]
    for root in roots:
        for dirpath, _, filenames in os.walk(root):
            for filename in filenames:
                if filename.endswith(".py"):
                    python_paths.append(os.path.join(dirpath, filename))
    return python_paths


def test_ensure_copyright():
    """Check that all non-empty Python files have copyright somewhere in the first 5 lines."""
    issues = []
    regex = re.compile(r"# Copyright \d{4}(-\d{4})? Canonical Ltd.$")
    for filepath in get_python_filepaths():
        if pathlib.Path(filepath).stat().st_size == 0:
            continue
        if filepath.endswith("charmcraft/_version.py") or filepath.endswith(
            "charmcraft\\_version.py"
        ):
            continue

        with open(filepath, encoding="utf8") as fh:
            for line in itertools.islice(fh, 5):
                if regex.match(line):
                    break
            else:
                issues.append(filepath)
    if issues:
        msg = "Please add copyright headers to the following files:\n" + "\n".join(issues)
        pytest.fail(msg, pytrace=False)



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for application class."""
import textwrap
from unittest import mock

import craft_application
import craft_cli.pytest_plugin
import pyfakefs.fake_filesystem
import pytest
from craft_application import util

from charmcraft import application, errors, services
from charmcraft.application.main import PRIME_BEHAVIOUR_CHANGE_MESSAGE


@pytest.mark.parametrize(
    ("charmcraft_dict", "metadata_yaml"),
    [
        (
            {},
            textwrap.dedent(
                """\
                name: test-charm
                summary: A test charm
                description: A charm for testing!"""
            ),
        ),
        (
            {"name": "test-charm"},
            textwrap.dedent(
                """\
                summary: A test charm
                description: A charm for testing!"""
            ),
        ),
        (
            {"name": "test-charm", "summary": "A test charm"},
            textwrap.dedent(
                """\
                description: A charm for testing!"""
            ),
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
            },
            textwrap.dedent(
                """\
                something-else: yes
                """
            ),
        ),
    ],
)
@pytest.mark.parametrize(
    "expected",
    [{"name": "test-charm", "summary": "A test charm", "description": "A charm for testing!"}],
)
def test_extra_yaml_transform_success(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    service_factory,
    charmcraft_dict,
    metadata_yaml,
    expected,
):
    """Test that _extra_yaml_transform correctly transforms the data."""
    fs.create_file("metadata.yaml", contents=metadata_yaml)
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    actual = app._extra_yaml_transform(charmcraft_dict, build_on="amd64", build_for=None)

    assert actual == expected


@pytest.mark.parametrize(
    ("charmcraft_dict", "metadata_yaml", "message"),
    [
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
            },
            "",
            "Invalid file: 'metadata.yaml'",
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
            },
            textwrap.dedent(
                """\
                name: test-charm
                summary: A test charm
                description: A charm for testing!"""
            ),
            "Fields in charmcraft.yaml cannot be duplicated in metadata.yaml",
        ),
    ],
)
def test_extra_yaml_transform_failure(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    service_factory,
    charmcraft_dict,
    metadata_yaml,
    message,
):
    fs.create_file("metadata.yaml", contents=metadata_yaml)
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    with pytest.raises(errors.CraftError) as exc_info:
        app._extra_yaml_transform(charmcraft_dict, build_for=None, build_on="amd64")

    assert exc_info.value.args[0] == message


@pytest.mark.parametrize(
    ("charmcraft_dict"),
    [
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"charm": {"prime": ["something"]}},
            }
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"bundle": {"prime": ["something"]}},
            }
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"reactive": {"prime": ["something"]}},
            }
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"other_name": {"plugin": "charm", "prime": ["something"]}},
            }
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"other_name": {"plugin": "bundle", "prime": ["something"]}},
            }
        ),
        (
            {
                "name": "test-charm",
                "summary": "A test charm",
                "description": "A charm for testing!",
                "parts": {"other_name": {"plugin": "reactive", "prime": ["something"]}},
            }
        ),
    ],
)
def test_deprecated_prime_warning(
    emitter,
    service_factory,
    charmcraft_dict,
):
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    app._extra_yaml_transform(charmcraft_dict, build_for=None, build_on="amd64")

    emitter.assert_progress(PRIME_BEHAVIOUR_CHANGE_MESSAGE, permanent=True)


@pytest.mark.parametrize(
    "base_charm",
    [
        {
            "name": "test-charm",
            "summary": "A test charm",
            "description": "A charm for testing!",
        },
    ],
)
@pytest.mark.parametrize(
    ("parts"),
    [
        pytest.param({}, id="no-parts"),
        pytest.param(
            {
                "parts": {"charm": {}},
            },
            id="named-charm",
        ),
        pytest.param({"parts": {"my-part": {"plugin": "charm"}}}, id="charm-plugin"),
        pytest.param(
            {
                "parts": {"reactive": {}},
            },
            id="named-reactive",
        ),
        pytest.param({"parts": {"my-part": {"plugin": "reactive"}}}, id="reactive-plugin"),
        pytest.param(
            {
                "parts": {"bundle": {}},
            },
            id="named-bundle",
        ),
        pytest.param({"parts": {"my-part": {"plugin": "bundle"}}}, id="bundle-plugin"),
    ],
)
def test_deprecated_prime_warning_not_raised(
    emitter,
    service_factory: services.CharmcraftServiceFactory,
    base_charm: dict[str, str],
    parts: dict[str, dict[str, str]],
):
    charmcraft_dict = base_charm | parts
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    app._extra_yaml_transform(charmcraft_dict, build_for=None, build_on="amd64")

    with pytest.raises(AssertionError, match="^Expected call"):
        emitter.assert_progress(PRIME_BEHAVIOUR_CHANGE_MESSAGE, permanent=True)


@pytest.mark.parametrize(
    "charm_yaml",
    [
        {
            "name": "test-charm",
            "summary": "A test charm",
            "description": "A charm for testing!",
            "parts": {"charm": {"prime": ["something"]}},
        },
    ],
)
def test_deprecated_prime_warning_not_raised_in_managed_mode(
    monkeypatch, emitter, service_factory: services.CharmcraftServiceFactory, charm_yaml
):
    monkeypatch.setenv("CRAFT_MANAGED_MODE", "1")

    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    app._extra_yaml_transform(charm_yaml, build_for=None, build_on="riscv64")


@pytest.mark.parametrize(
    "build_for",
    [
        "amd64-arm64",
        "s390x-ppc64el-riscv64",
    ],
)
def test_expand_environment_multi_arch(
    monkeypatch: pytest.MonkeyPatch,
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    service_factory: services.CharmcraftServiceFactory,
    build_for,
) -> None:
    mock_parent_expand_environment = mock.Mock()
    monkeypatch.setattr(
        craft_application.Application, "_expand_environment", mock_parent_expand_environment
    )
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)

    app._expand_environment({}, build_for)

    emitter.assert_debug(
        "Expanding environment variables with the host architecture "
        f"{util.get_host_architecture()!r} as the build-for architecture "
        "because multiple run-on architectures were specified."
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft metadata pydantic model."""

from typing import TYPE_CHECKING, Any

import pydantic
from craft_application import models
from craft_application.models import base
from craft_cli import CraftError
from typing_extensions import Self, override

from charmcraft import const

if TYPE_CHECKING:
    from charmcraft.models.project import Bundle, Charm
else:
    Charm = Bundle = None


class CharmMetadata(models.BaseMetadata):
    """A charm's metadata.yaml file.

    This model represents the metadata.yaml file that gets placed in the root
    of a charm.
    """

    model_config = pydantic.ConfigDict(
        validate_assignment=True,
        extra="ignore",
        populate_by_name=True,
        alias_generator=base.alias_generator,
    )

    name: models.ProjectName
    display_name: models.ProjectTitle | None = None
    summary: pydantic.StrictStr
    description: pydantic.StrictStr
    maintainers: list[pydantic.StrictStr] | None = None
    assumes: list[str | dict[str, list | dict]] | None = None
    charm_user: str | None = None
    containers: dict[str, Any] | None = None
    devices: dict[str, Any] | None = None
    docs: pydantic.AnyHttpUrl | None = None
    extra_bindings: dict[str, Any] | None = None
    issues: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None
    peers: dict[str, Any] | None = None
    provides: dict[str, Any] | None = None
    requires: dict[str, Any] | None = None
    resources: dict[str, Any] | None = None
    source: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None
    storage: dict[str, Any] | None = None
    subordinate: bool | None = None
    terms: list[pydantic.StrictStr] | None = None
    website: pydantic.AnyHttpUrl | list[pydantic.AnyHttpUrl] | None = None

    @classmethod
    def from_charm(cls, charm: Charm) -> Self:
        """Turn a populated charm model into a metadata model.

        Performs the necessary renaming and reorganisation.
        """
        charm_dict = charm.model_dump(
            exclude_none=True,
            by_alias=True,
            exclude_defaults=False,
        )

        # Flatten links and match to the appropriate metadata.yaml schema
        if charm.links is not None:
            links = charm.links.marshal()
            if "documentation" in links:
                charm_dict["docs"] = links.pop("documentation")
            if "contact" in links:
                contact = links.pop("contact")
                if isinstance(contact, str):
                    contact = [contact]
                charm_dict["maintainers"] = contact
            charm_dict.update(links)

        if "title" in charm_dict:
            charm_dict["display-name"] = charm_dict.pop("title")

        return cls.model_validate(charm_dict)


class CharmMetadataLegacy(CharmMetadata):
    """Object representing LEGACY charm metadata.yaml contents.

    This model only supports the legacy charm metadata.yaml format for compatibility.
    New metadata defined in charmcraft.yaml is handled by the CharmcraftProject models.

    specs: https://juju.is/docs/sdk/metadata-yaml
    """

    # These are looser in the metadata.yaml schema than charmcraft requires.
    name: pydantic.StrictStr  # type: ignore[assignment]
    summary: pydantic.StrictStr  # type: ignore[assignment]
    description: pydantic.StrictStr
    display_name: pydantic.StrictStr | None = None  # type: ignore[assignment]

    @override
    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> Self:
        """Unmarshal object with necessary translations and error handling.

        :returns: valid CharmMetadataLegacy object.

        :raises CraftError: On failure to unmarshal object.
        """
        # convert undocumented "maintainer" to documented "maintainers"
        if "maintainer" in data and "maintainers" in data:
            raise CraftError(
                f"Cannot specify both 'maintainer' and 'maintainers' in {const.METADATA_FILENAME}"
            )

        if "maintainer" in data:
            data["maintainers"] = [data["maintainer"]]
            del data["maintainer"]

        return cls.model_validate(data)


class BundleMetadata(models.BaseMetadata):
    """metadata.yaml for a bundle zip."""

    name: models.ProjectName | None = None
    description: pydantic.StrictStr | None = None

    @classmethod
    def from_bundle(cls, bundle: Bundle) -> Self:
        """Turn a populated bundle model into a metadata.yaml model."""
        bundle_dict = bundle.marshal()
        if "bundle" in bundle_dict:
            return cls.model_validate(bundle_dict["bundle"])
        del bundle_dict["type"]
        return cls.model_validate(bundle_dict)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""A wrapper around Skopeo."""

import io
import json
import pathlib
import shutil
import subprocess
from collections.abc import Sequence
from typing import Any, cast, overload

from charmcraft import errors


class Skopeo:
    """A class for interacting with skopeo."""

    def __init__(
        self,
        *,
        skopeo_path: str = "",
        insecure_policy: bool = False,
        arch: str | None = None,
        os: str | None = None,
        tmpdir: pathlib.Path | None = None,
        debug: bool = False,
    ) -> None:
        if skopeo_path:
            self._skopeo = skopeo_path
        else:
            self._skopeo = cast(str, shutil.which("skopeo"))
            if not self._skopeo:
                raise RuntimeError("Cannot find a skopeo executable.")
        self._insecure_policy = insecure_policy
        self.arch = arch
        self.os = os
        if tmpdir:
            tmpdir.mkdir(parents=True, exist_ok=True)
        self._tmpdir = tmpdir
        self._debug = debug

        self._run_skopeo([self._skopeo, "--version"], capture_output=True, text=True)

    def get_global_command(self) -> list[str]:
        """Prepare the global skopeo options."""
        command = [self._skopeo]
        if self._insecure_policy:
            command.append("--insecure-policy")
        if self.arch:
            command.extend(["--override-arch", self.arch])
        if self.os:
            command.extend(["--override-os", self.os])
        if self._tmpdir:
            command.extend(["--tmpdir", str(self._tmpdir)])
        if self._debug:
            command.append("--debug")
        return command

    def _run_skopeo(self, command: Sequence[str], **kwargs) -> subprocess.CompletedProcess:
        """Run skopeo, converting the error message if necessary."""
        try:
            return subprocess.run(command, check=True, **kwargs)
        except subprocess.CalledProcessError as exc:
            raise errors.SubprocessError.from_subprocess(exc) from exc

    def copy(
        self,
        source_image: str,
        destination_image: str,
        *,
        all_images: bool = False,
        preserve_digests: bool = False,
        source_username: str | None = None,
        source_password: str | None = None,
        dest_username: str | None = None,
        dest_password: str | None = None,
        stdout: io.FileIO | int | None = None,
        stderr: io.FileIO | int | None = None,
    ) -> subprocess.CompletedProcess:
        """Copy an OCI image using Skopeo."""
        command = [
            *self.get_global_command(),
            "copy",
        ]
        if all_images:
            command.append("--all")
        if preserve_digests:
            command.append("--preserve-digests")
        if source_username and source_password:
            command.extend(["--src-creds", f"{source_username}:{source_password}"])
        elif source_username:
            command.extend(["--src-creds", source_username])
        elif source_password:
            command.extend(["--src-password", source_password])
        if dest_username and dest_password:
            command.extend(["--dest-creds", f"{dest_username}:{dest_password}"])
        elif dest_username:
            command.extend(["--dest-creds", dest_username])
        elif dest_password:
            command.extend(["--dest-password", dest_password])

        command.extend([source_image, destination_image])

        if stdout or stderr:
            return self._run_skopeo(command, stdout=stdout, stderr=stderr, text=True)
        return self._run_skopeo(command, capture_output=True, text=True)

    @overload
    def inspect(
        self, image: str, *, format_template: None = None, raw: bool = False, tags: bool = True
    ) -> dict[str, Any]: ...
    @overload
    def inspect(
        self, image: str, *, format_template: str, raw: bool = False, tags: bool = True
    ) -> str: ...
    def inspect(
        self,
        image: str,
        *,
        format_template: str | None = None,
        raw: bool = False,
        tags: bool = True,
    ) -> dict[str, Any] | str:
        """Inspect an image."""
        command = [*self.get_global_command(), "inspect"]
        if format_template is not None:
            command.extend(["--format", format_template])
        if raw:
            command.append("--raw")
        if not tags:
            command.append("--no-tags")

        command.append(image)

        result = self._run_skopeo(command, capture_output=True, text=True)

        if format_template is None:
            return json.loads(result.stdout)
        return result.stdout



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft pydantic models."""

from . import actions, config, metadata
from .charmcraft import Base
from .lint import CheckResult, CheckType, LintResult, ResultLevel
from .manifest import Attribute, Manifest
from .metadata import BundleMetadata, CharmMetadata, CharmMetadataLegacy
from .project import (
    CharmBuildInfo,
    CharmcraftBuildPlanner,
    CharmLib,
    CharmcraftProject,
    BasesCharm,
    PlatformCharm,
    Charm,
    Bundle,
)

__all__ = [
    "actions",
    "config",
    "metadata",
    "Base",
    "CheckResult",
    "CheckType",
    "LintResult",
    "ResultLevel",
    "Attribute",
    "Manifest",
    "Bundle",
    "BasesCharm",
    "PlatformCharm",
    "Charm",
    "CharmBuildInfo",
    "CharmcraftBuildPlanner",
    "CharmcraftProject",
    "CharmLib",
    "BundleMetadata",
    "CharmMetadata",
    "CharmMetadataLegacy",
]



# Copyright 2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Script to visualize information produced by Charmcraft with --measure.

It needs the following dependencies, for which you need to set up a
virtualenv or just use fades:

    matplotlib
"""

import argparse
import itertools
import json
import time

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

Y_STEPS = 10
COLORS = itertools.cycle(mcolors.TABLEAU_COLORS)


def main(filepath):
    """Main entry point."""

    with open(filepath, "rt") as fh:
        results = json.load(fh)

    baseline = results.pop("__meta__")["baseline"]

    by_parent = {}
    for mid, info in results.items():
        by_parent.setdefault(info["parent"], []).append(mid)

    layers = []
    locating = [None]
    while locating:
        layer = []
        for parent in locating:
            for mid in by_parent.get(parent, []):
                layer.append(mid)

        if layer:
            layers.append(layer)
        locating = layer

    # get the measurement limits to only show texts when relevant or decide its angle
    max_t = max(measurement["tend"] for measurement in results.values())
    wide_metric = max_t * 0.05  # more than this, show it horizontal
    ignorable_metric = max_t * 0.005  # less than this, don't show it

    fig, ax = plt.subplots()

    layer_y_pos = (len(layers) - 1) * Y_STEPS
    layer_height = Y_STEPS * 0.99
    for layer in layers:
        measurements = [results[mid] for mid in layer]

        # the bars
        x_ranges = []
        for measurement in measurements:
            x_ranges.append((measurement["tstart"], measurement["tend"] - measurement["tstart"]))
        colors = [next(COLORS) for _ in x_ranges]
        ax.broken_barh(x_ranges, (layer_y_pos, layer_height), facecolors=colors)

        # the texts
        text_y_pos = layer_y_pos + Y_STEPS / 2
        for measurement in measurements:
            tdelta = measurement["tend"] - measurement["tstart"]
            if tdelta < ignorable_metric:
                continue
            if tdelta < wide_metric:
                rotation = 90
            else:
                rotation = 0
            text = measurement["msg"]
            if measurement["extra"]:
                text += "\n" + str(measurement["extra"])
            bar_center = (measurement["tend"] + measurement["tstart"]) / 2
            ax.text(
                x=bar_center,
                y=text_y_pos,
                s=text,
                ha="center",
                va="center",
                color="black",
                rotation=rotation,
            )

        layer_y_pos -= Y_STEPS

    ax.set_yticklabels([])
    ax.set_xlabel("seconds")
    plt.title(f"Measurements took at {time.ctime(baseline)}")
    plt.xlim([0, max_t])
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("filepath", help="The produced file after using --measure in Charmcraft")
    args = parser.parse_args()
    main(args.filepath)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Model for output charm's manifest.yaml file."""
from typing import Any, Literal

from craft_application import models

import charmcraft
from charmcraft.models.charmcraft import Base


class Attribute(models.BaseMetadata):
    """An attribute as a linter result."""

    name: str
    result: str


class Manifest(models.BaseMetadata):
    """A manifest.yaml file for Juju.

    See: https://juju.is/docs/sdk/manifest-yaml
    """

    charmcraft_version: str = charmcraft.__version__
    charmcraft_started_at: str
    bases: list[Base] | None
    analysis: dict[Literal["attributes"], list[Attribute]] = {"attributes": []}
    image_info: Any | None = None



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from typing import Any

import pytest
from overrides import override

from charmcraft import const, errors, extensions
from charmcraft.extensions.extension import Extension


class FakeExtension(Extension):
    """A fake test Extension"""

    name = "fake-extension"
    bases = [("ubuntu", "22.04")]

    @classmethod
    def get_supported_bases(cls) -> list[tuple[str, ...]]:
        """Return a list of tuple of supported bases."""
        return cls.bases

    @staticmethod
    def is_experimental(_base: tuple[str, ...] | None) -> bool:
        """Return whether or not this extension is unstable for given base."""
        return False

    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {}

    def get_part_snippet(self) -> dict[str, Any]:
        """Return the part snippet to apply to existing parts."""
        return {}

    def get_parts_snippet(self) -> dict[str, Any]:
        """Return the parts to add to parts."""
        return {}


class ExperimentalExtension(FakeExtension):
    """A fake test Extension that is experimental"""

    name = "experimental-extension"
    bases = [("ubuntu", "22.04")]

    @staticmethod
    def is_experimental(_base: str | None) -> bool:
        return True


class InvalidPartExtension(FakeExtension):
    """A fake test Extension that has invalid parts snippet"""

    name = "invalid-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_parts_snippet(self) -> dict[str, Any]:
        return {"bad-name": {"plugin": "dump", "source": None}}


class FullExtension(FakeExtension):
    """A fake test Extension that has complete behavior"""

    name = "full-extension"
    bases = [("ubuntu", "22.04")]

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to apply."""
        return {
            "terms": ["https://example.com/terms", "https://example.com/terms2"],
        }

    @override
    def get_part_snippet(self) -> dict[str, Any]:
        """Return the part snippet to apply to existing parts."""
        return {"stage-packages": ["new-package-1"]}

    @override
    def get_parts_snippet(self) -> dict[str, Any]:
        """Return the parts to add to parts."""
        return {"full-extension/new-part": {"plugin": "nil", "source": None}}


@pytest.fixture
def fake_extensions(stub_extensions):
    extensions.register(FakeExtension.name, FakeExtension)
    extensions.register(ExperimentalExtension.name, ExperimentalExtension)
    extensions.register(InvalidPartExtension.name, InvalidPartExtension)
    extensions.register(FullExtension.name, FullExtension)


def test_experimental_with_env(fake_extensions, tmp_path, monkeypatch):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [ExperimentalExtension.name],
    }
    monkeypatch.setenv(const.EXPERIMENTAL_EXTENSIONS_ENV_VAR, "1")
    project_root = tmp_path
    extensions.apply_extensions(project_root, charmcraft_config)


def test_experimental_no_env(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [
            {
                "build-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
                "run-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
            }
        ],
        "extensions": [ExperimentalExtension.name],
    }
    with pytest.raises(errors.ExtensionError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    expected_message = f"Extension is experimental: '{ExperimentalExtension.name}'"
    assert str(exc.value) == expected_message


def test_wrong_base(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [
            {
                "build-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
                "run-on": [{"name": "ubuntu", "channel": "20.04", "architectures": ["amd64"]}],
            }
        ],
        "extensions": [FakeExtension.name],
    }
    with pytest.raises(errors.ExtensionError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    expected_message = (
        f"Extension '{FakeExtension.name}' does not support base: ('ubuntu', '20.04')"
    )
    assert str(exc.value) == expected_message


def test_invalid_parts(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [InvalidPartExtension.name],
    }

    with pytest.raises(ValueError) as exc:
        extensions.apply_extensions(tmp_path, charmcraft_config)

    assert "Extension has invalid part names" in str(exc.value)


def test_apply_extensions(fake_extensions, tmp_path):
    charmcraft_config = {
        "type": "charm",
        "name": "test-charm-name-from-charmcraft-yaml",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": [FullExtension.name],
        "parts": {"my-part": {"plugin": "nil", "source": None, "stage-packages": ["old-package"]}},
    }

    applied = extensions.apply_extensions(tmp_path, charmcraft_config)

    # Part snippet extends the existing part
    parts = applied["parts"]
    assert parts["my-part"]["stage-packages"] == [
        "new-package-1",
        "old-package",
    ]

    # New part
    assert parts[f"{FullExtension.name}/new-part"] == {"plugin": "nil", "source": None}



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for init command."""
import argparse
import contextlib
import os
import pathlib
import re
import shutil
import subprocess
import sys
from unittest import mock

import pydocstyle
import pytest
import pytest_check

import charmcraft
from charmcraft import errors
from charmcraft.application import commands
from charmcraft.utils import S_IXALL

with contextlib.suppress(ImportError):
    import pwd

BASIC_INIT_FILES = frozenset(
    pathlib.Path(p)
    for p in (
        ".gitignore",
        "charmcraft.yaml",
        "CONTRIBUTING.md",
        "LICENSE",
        "pyproject.toml",
        "README.md",
        "requirements.txt",
        "src",
        "src/charm.py",
        "tests",
        "tests/integration",
        "tests/integration/test_charm.py",
        "tests/unit",
        "tests/unit/test_charm.py",
        "tox.ini",
    )
)
BASIC_INIT_FILES_WITH_SPREAD = frozenset(
    list(BASIC_INIT_FILES)
    + [
        pathlib.Path(p)
        for p in (
            "spread.yaml",
            "tests/spread",
            "tests/spread/lib",
            "tests/spread/lib/cloud-config.yaml",
            "tests/spread/lib/test-helpers.sh",
            "tests/spread/general",
            "tests/spread/general/integration",
            "tests/spread/general/integration/task.yaml",
        )
    ]
)
BASIC_INIT_FILES_WITH_TOOLS = frozenset(
    list(BASIC_INIT_FILES_WITH_SPREAD)
    + [
        pathlib.Path(p)
        for p in (
            "tests/spread/lib",
            "tests/spread/lib/tools",
            "tests/spread/lib/tools/retry",
        )
    ]
)
UNKNOWN_AUTHOR_REGEX = re.compile(
    r"^Unable to automatically determine author's name, specify it with --author$"
)
BAD_CHARM_NAME_REGEX = re.compile(
    r" is not a valid charm name. The name must start with a lowercase letter and contain only alphanumeric characters and hyphens.$",
)
VALID_AUTHORS = [
    pytest.param("Author McAuthorFace", id="ascii-author"),
    pytest.param("فلانة الفلانية", id="non-ascii-author"),
]


@pytest.fixture
def init_command():
    return commands.InitCommand({"app": charmcraft.application.APP_METADATA, "services": None})


def create_namespace(
    *,
    name="my-charm",
    author="J Doe",
    force=False,
    profile=commands.init.DEFAULT_PROFILE,
    project_dir: pathlib.Path | None = None,
):
    """Helper to create a valid namespace."""
    if project_dir is None:
        project_dir = pathlib.Path.cwd()
    return argparse.Namespace(
        name=name,
        author=author,
        force=force,
        profile=profile,
        project_dir=project_dir,
    )


@pytest.mark.parametrize(
    ("profile", "expected_files"),
    [
        pytest.param("simple", BASIC_INIT_FILES_WITH_TOOLS, id="simple"),
        pytest.param("machine", BASIC_INIT_FILES_WITH_SPREAD, id="machine"),
        pytest.param("kubernetes", BASIC_INIT_FILES_WITH_TOOLS, id="kubernetes"),
    ],
)
@pytest.mark.parametrize("charm_name", ["my-charm", "charm123"])
@pytest.mark.parametrize("author", VALID_AUTHORS)
def test_files_created_correct(
    new_path,
    init_command,
    profile: str,
    expected_files: set[pathlib.Path],
    charm_name,
    author,
):
    params = create_namespace(name=charm_name, author=author, profile=profile)
    init_command.run(params)

    actual_files = {p.relative_to(new_path) for p in new_path.rglob("*")}

    # Note: we need to specify the encoding here because Windows defaults ta CP-1252.
    charmcraft_yaml = (new_path / "charmcraft.yaml").read_text(encoding="utf-8")
    tox_ini = (new_path / "tox.ini").read_text(encoding="utf-8")

    pytest_check.equal(actual_files, expected_files)
    pytest_check.is_true(re.search(rf"^name: {charm_name}$", charmcraft_yaml, re.MULTILINE))
    pytest_check.is_true(re.search(rf"^# Copyright \d+ {author}", tox_ini))


def test_force(new_path, init_command):
    tmp_file = new_path / "README.md"
    with tmp_file.open("w") as f:
        f.write("This is a nonsense readme")

    init_command.run(create_namespace(force=True))

    # Check that init ran
    assert (new_path / "LICENSE").exists()

    # Check that init did not overwrite files
    with tmp_file.open("r") as f:
        assert f.read() == "This is a nonsense readme"


@pytest.mark.parametrize("name", [None, 0, "1234", "yolo swag", "camelCase"])
def test_bad_name(monkeypatch, new_path, init_command, name):
    with pytest.raises(errors.CraftError, match=BAD_CHARM_NAME_REGEX):
        init_command.run(create_namespace(name=name))


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
@pytest.mark.parametrize("author", VALID_AUTHORS)
def test_gecos_valid_author(monkeypatch, new_path, init_command, author):
    monkeypatch.setattr(
        pwd,
        "getpwuid",
        mock.Mock(
            return_value=pwd.struct_passwd(
                ("user", "pass", 1, 1, f"{author},,,", "homedir", "shell")
            )
        ),
    )

    init_command.run(create_namespace(author=None))

    pytest_check.is_true(
        re.search(rf"^# Copyright \d+ {author}", (new_path / "tox.ini").read_text())
    )


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
@pytest.mark.parametrize(
    ("mock_getpwuid", "error_msg"),
    [
        pytest.param(
            mock.Mock(side_effect=KeyError("no user")),
            UNKNOWN_AUTHOR_REGEX,
            id="user-doesnt-exist",
        ),
    ],
)
def test_gecos_user_not_found(monkeypatch, new_path, init_command, mock_getpwuid, error_msg):
    monkeypatch.setattr(pwd, "getpwuid", mock_getpwuid)

    with pytest.raises(errors.CraftError, match=error_msg):
        init_command.run(create_namespace(author=None))


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
def test_gecos_user_has_no_name(monkeypatch, new_path, init_command):
    mock_getpwuid = mock.Mock(
        return_value=pwd.struct_passwd(("user", "pass", 1, 1, "", "dir", "shell"))
    )
    monkeypatch.setattr(pwd, "getpwuid", mock_getpwuid)

    with pytest.raises(errors.CraftError, match=UNKNOWN_AUTHOR_REGEX):
        init_command.run(create_namespace(author=None))


@pytest.mark.parametrize(
    "subdir",
    [
        "somedir",
        "some/dir",
        "a/really/deep/path/with_parents/all/created",
        pytest.param(
            "/tmp/test_charm_dir-absolute_directory_path",
            marks=pytest.mark.skipif(os.name != "posix", reason="This is a posix path"),
        ),
    ],
)
@pytest.mark.parametrize("expected_files", [BASIC_INIT_FILES_WITH_TOOLS])
def test_create_directory(new_path, init_command, subdir, expected_files):
    init_dir = new_path / subdir

    try:
        init_command.run(create_namespace(project_dir=pathlib.Path(subdir)))

        actual_files = {p.relative_to(init_dir) for p in init_dir.rglob("*")}

        assert actual_files == expected_files

    finally:
        shutil.rmtree(init_dir)


@pytest.mark.skipif(os.name != "posix", reason="Checking posix executable bit.")
def test_executable_set(new_path, init_command):
    init_command.run(create_namespace())

    for path in new_path.rglob(".py"):
        assert path.stat().st_mode & S_IXALL == S_IXALL


@pytest.mark.slow
@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.skipif(
    bool(os.getenv("RUNNING_TOX")) and sys.version_info < (3, 11),
    reason="does not work inside tox in Python3.10 and below",
)
@pytest.mark.parametrize("profile", list(commands.init.PROFILES))
def test_tox_success(new_path, init_command, profile):
    # fix the PYTHONPATH and PATH so the tests in the initted environment use our own
    # virtualenv libs and bins (if any), as they need them, but we're not creating a
    # venv for the local tests (note that for CI doesn't use a venv)
    env = os.environ.copy()
    env_paths = [p for p in sys.path if "env/lib/python" in p]
    if env_paths:
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += ":" + ":".join(env_paths)
        else:
            env["PYTHONPATH"] = ":".join(env_paths)
        for path in env_paths:
            bin_path = path[: path.index("env/lib/python")] + "env/bin"
            env["PATH"] = bin_path + ":" + env["PATH"]

    init_command.run(create_namespace(profile=profile))

    if not (new_path / "tox.ini").exists():
        pytest.skip("init template doesn't contain tox.ini file")

    result = subprocess.run(
        ["tox", "-v"],
        cwd=new_path,
        env=env,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    )
    assert result.returncode == 0, "Tox run failed:\n" + result.stdout


@pytest.mark.parametrize("profile", list(commands.init.PROFILES))
def test_pep257(new_path, init_command, profile):
    to_ignore = {
        "D105",  # Missing docstring in magic method
        "D107",  # Missing docstring in __init__
    }
    to_include = pydocstyle.violations.conventions.pep257 - to_ignore

    init_command.run(create_namespace(profile=profile))

    python_paths = (str(path) for path in new_path.rglob("*.py"))
    python_paths = (path for path in python_paths if "tests" not in path)
    errors = list(pydocstyle.check(python_paths, select=to_include))

    if errors:
        report = [f"Please fix files as suggested by pydocstyle ({len(errors):d} issues):"]
        report.extend(str(e) for e in errors)
        msg = "\n".join(report)
        pytest.fail(msg, pytrace=False)



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Gunicorn based extensions."""

from typing import Any

from overrides import override

from ..errors import ExtensionError
from .extension import Extension


class _AppBase(Extension):
    """A base class for 12-factor applications."""

    _CHARM_LIBS = [
        {"lib": "traefik_k8s.ingress", "version": "2"},
        {"lib": "observability_libs.juju_topology", "version": "0"},
        {"lib": "grafana_k8s.grafana_dashboard", "version": "0"},
        {"lib": "loki_k8s.loki_push_api", "version": "0"},
        {"lib": "data_platform_libs.data_interfaces", "version": "0"},
        {"lib": "prometheus_k8s.prometheus_scrape", "version": "0"},
        {"lib": "redis_k8s.redis", "version": "0"},
        {"lib": "data_platform_libs.s3", "version": "0"},
        {"lib": "saml_integrator.saml", "version": "0"},
    ]

    @staticmethod
    @override
    def get_supported_bases() -> list[tuple[str, str]]:
        """Return supported bases."""
        return [("ubuntu", "22.04")]

    @staticmethod
    @override
    def is_experimental(base: tuple[str, ...] | None) -> bool:  # noqa: ARG004
        """Check if the extension is in an experimental state."""
        return True

    framework: str
    actions: dict = {
        "rotate-secret-key": {
            "description": "Rotate the secret key. Users will be forced to log in again. This might be useful if a security breach occurs."
        }
    }

    options: dict

    def _get_nested(self, obj: dict, path: str) -> dict:
        """Get a nested object using a path (a dot-separated list of keys)."""
        for key in path.split("."):
            obj = obj.get(key, {})
        return obj

    def _check_input(self) -> None:
        """Check if the extension is applicable for user input charmcraft project file."""
        charm_type = self.yaml_data.get("type")
        if charm_type != "charm":
            raise ExtensionError(
                f"the '{self.framework}-framework' extension is incompatible with "
                f"type {charm_type!r}"
            )
        parts = self.yaml_data.get("parts")
        if parts and "charm" in parts:
            raise ExtensionError(
                f"the '{self.framework}-framework' extension is incompatible with "
                f"customized charm part"
            )
        incompatible_fields = {"devices", "extra-bindings", "storage"} & self.yaml_data.keys()
        if incompatible_fields:
            raise ExtensionError(
                f"the '{self.framework}-framework' extension is incompatible with the provided "
                f"field(s): {', '.join(sorted(incompatible_fields))}"
            )
        root_snippet = self._get_root_snippet()
        for protected in ("assumes", "containers", "resources", "peers"):
            if (
                protected in self.yaml_data
                and self.yaml_data[protected] != root_snippet[protected]
            ):
                raise ExtensionError(
                    f"{protected!r} in charmcraft.yaml conflicts with a reserved field "
                    f"in the {self.framework}-framework extension, please remove it."
                )
        for merging in ("actions", "requires", "provides", "config.options"):
            user_provided: dict[str, Any] = self._get_nested(self.yaml_data, merging)
            if not user_provided:
                continue
            overlap = user_provided.keys() & self._get_nested(root_snippet, merging).keys()
            if overlap:
                raise ExtensionError(
                    f"overlapping keys {overlap} in {merging} of charmcraft.yaml "
                    f"which conflict with the {self.framework}-framework extension, "
                    "please rename or remove it"
                )
        for config in self._get_nested(self.yaml_data, "config.options"):
            for reserved_config_prefix in ("webserver-", f"{self.framework}-"):
                if config.startswith(reserved_config_prefix):
                    raise ExtensionError(
                        f"config.options {config!r} starts with {self.framework}-framework"
                        f" reserved configuration prefix {reserved_config_prefix!r}, "
                        "please rename or remove it"
                    )

    def _get_root_snippet(self) -> dict[str, Any]:
        """Return the root snippet to be merged into the user charmcraft.yaml.

        This method differs from get_root_snippet because it doesn't perform any check.
        """
        return {
            "assumes": ["k8s-api"],
            "containers": {
                self.get_container_name(): {"resource": self.get_image_name()},
            },
            "resources": {
                self.get_image_name(): {
                    "type": "oci-image",
                    "description": f"{self.framework} application image.",
                },
            },
            "charm-libs": self._CHARM_LIBS,
            "peers": {"secret-storage": {"interface": "secret-storage"}},
            "actions": self.actions,
            "requires": {
                "logging": {"interface": "loki_push_api"},
                "ingress": {"interface": "ingress", "limit": 1},
            },
            "provides": {
                "metrics-endpoint": {"interface": "prometheus_scrape"},
                "grafana-dashboard": {"interface": "grafana_dashboard"},
            },
            "config": {"options": self.options},
            "parts": {
                "charm": {
                    "plugin": "charm",
                    "source": ".",
                    "build-snaps": ["rustup"],  # Needed to build pydantic.
                    "override-build": "rustup default stable\ncraftctl default",
                }
            },
        }

    @override
    def get_root_snippet(self) -> dict[str, Any]:
        """Fill in some required root components."""
        self._check_input()
        return self._get_root_snippet()

    @override
    def get_part_snippet(self) -> dict[str, Any]:
        """Return the part snippet to apply to existing parts."""
        return {}

    @override
    def get_parts_snippet(self) -> dict[str, Any]:
        """Return the parts to add to parts."""
        return {}

    def get_container_name(self) -> str:
        """Return name of the container for the app image."""
        return f"{self.framework}-app"

    def get_image_name(self) -> str:
        """Return name of the app image."""
        return f"{self.framework}-app-image"


GUNICORN_WEBSERVER_OPTIONS = {
    "webserver-keepalive": {
        "type": "int",
        "description": "Time in seconds for webserver to wait for requests on a Keep-Alive connection.",
    },
    "webserver-threads": {
        "type": "int",
        "description": "Run each webserver worker with the specified number of threads.",
    },
    "webserver-timeout": {
        "type": "int",
        "description": "Time in seconds to kill and restart silent webserver workers.",
    },
    "webserver-workers": {
        "type": "int",
        "description": "The number of webserver worker processes for handling requests.",
    },
}


class FlaskFramework(_AppBase):
    """Extension for 12-factor Flask applications."""

    framework = "flask"
    options = {
        **GUNICORN_WEBSERVER_OPTIONS,
        "flask-application-root": {
            "type": "string",
            "description": "Path in which the application / web server is mounted. This configuration will set the FLASK_APPLICATION_ROOT environment variable. Run `app.config.from_prefixed_env()` in your Flask application in order to receive this configuration.",
        },
        "flask-debug": {
            "type": "boolean",
            "description": "Whether Flask debug mode is enabled.",
        },
        "flask-env": {
            "type": "string",
            "description": "What environment the Flask app is running in, by default it's 'production'.",
        },
        "flask-permanent-session-lifetime": {
            "type": "int",
            "description": "Time in seconds for the cookie to expire in the Flask application permanent sessions. This configuration will set the FLASK_PERMANENT_SESSION_LIFETIME environment variable. Run `app.config.from_prefixed_env()` in your Flask application in order to receive this configuration.",
        },
        "flask-preferred-url-scheme": {
            "type": "string",
            "default": "HTTPS",
            "description": 'Scheme for generating external URLs when not in a request context in the Flask application. By default, it\'s "HTTPS". This configuration will set the FLASK_PREFERRED_URL_SCHEME environment variable. Run `app.config.from_prefixed_env()` in your Flask application in order to receive this configuration.',
        },
        "flask-secret-key": {
            "type": "string",
            "description": "The secret key used for securely signing the session cookie and for any other security related needs by your Flask application. This configuration will set the FLASK_SECRET_KEY environment variable. Run `app.config.from_prefixed_env()` in your Flask application in order to receive this configuration.",
        },
        "flask-session-cookie-secure": {
            "type": "boolean",
            "description": "Set the secure attribute in the Flask application cookies. This configuration will set the FLASK_SESSION_COOKIE_SECURE environment variable. Run `app.config.from_prefixed_env()` in your Flask application in order to receive this configuration.",
        },
    }

    @staticmethod
    @override
    def is_experimental(base: tuple[str, ...] | None) -> bool:  # noqa: ARG004
        """Check if the extension is in an experimental state."""
        return False


class DjangoFramework(_AppBase):
    """Extension for 12-factor Django applications."""

    framework = "django"
    actions = {
        **_AppBase.actions,
        "create-superuser": {
            "description": "Create a new Django superuser account.",
            "params": {"username": {"type": "string"}, "email": {"type": "string"}},
            "required": ["username", "email"],
        },
    }
    options = {
        **GUNICORN_WEBSERVER_OPTIONS,
        "django-debug": {
            "type": "boolean",
            "default": False,
            "description": "Whether Django debug mode is enabled.",
        },
        "django-secret-key": {
            "type": "string",
            "description": "The secret key used for securely signing the session cookie and for any other security related needs by your Django application. This configuration will set the DJANGO_SECRET_KEY environment variable.",
        },
        "django-allowed-hosts": {
            "type": "string",
            "description": "A comma-separated list of host/domain names that this Django site can serve. This configuration will set the DJANGO_ALLOWED_HOSTS environment variable with its content being a JSON encoded list.",
        },
    }


class GoFramework(_AppBase):
    """Extension for 12-factor Go applications."""

    framework = "go"
    options = {
        "app-port": {
            "type": "int",
            "default": 8080,
            "description": "Default port where the application will listen on.",
        },
        "metrics-port": {
            "type": "int",
            "default": 8080,
            "description": "Port where the prometheus metrics will be scraped.",
        },
        "metrics-path": {
            "type": "string",
            "default": "/metrics",
            "description": "Path where the prometheus metrics will be scraped.",
        },
        "app-secret-key": {
            "type": "string",
            "description": "Long secret you can use for sessions, csrf or any other thing where you need a random secret shared by all units",
        },
    }

    @staticmethod
    @override
    def get_supported_bases() -> list[tuple[str, str]]:
        """Return supported bases."""
        return [("ubuntu", "24.04")]

    @override
    def get_image_name(self) -> str:
        """Return name of the app image."""
        return "app-image"

    @override
    def get_container_name(self) -> str:
        """Return name of the container for the app image."""
        return "app"



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Build a charm remotely on Launchpad."""
import argparse
import os
import pathlib
import time
from collections.abc import Collection
from typing import Any, cast

from craft_application.commands import ExtensibleCommand
from craft_application.launchpad.models import Build, BuildState
from craft_application.remote.utils import get_build_id
from craft_cli import emit
from overrides import override  # pyright: ignore[reportUnknownVariableType]

from charmcraft import models, utils

OVERVIEW = """\
Command remote-build sends the current project to be built
remotely.  After the build is complete, packages for each
architecture are retrieved and will be available in the
local filesystem.

Interrupted remote builds can be resumed using the --recover
option, followed by the build number informed when the remote
build was originally dispatched. The current state of the
remote build for each architecture can be checked using the
--status option.

To set a timeout on the remote-build command, use the option
``--launchpad-timeout=<seconds>``. The timeout is local, so
the build on launchpad will continue even if the local instance
is interrupted or times out.
"""

_CONFIRMATION_PROMPT = (
    "All data sent to remote builders will be publicly available. "
    "Are you sure you want to continue?"
)


class RemoteBuild(ExtensibleCommand):
    """Run analysis on a built charm."""

    name = "remote-build"
    help_msg = "Build a charm remotely on Launchpad."
    overview = OVERVIEW
    always_load_project = True

    @override
    def _fill_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--recover", action="store_true", help="recover an interrupted build")
        parser.add_argument(
            "--launchpad-accept-public-upload",
            action="store_true",
            help="acknowledge that uploaded code will be publicly available.",
        )
        parser.add_argument(
            "--launchpad-timeout",
            type=int,
            default=0,
            metavar="<seconds>",
            help="Time in seconds to wait for launchpad to build.",
        )

    def _run(self, parsed_args: argparse.Namespace, **kwargs: Any) -> int | None:  # noqa: ANN401
        """Run the remote-build command.

        :param parsed_args: parsed argument namespace from craft_cli.

        :raises AcceptPublicUploadError: If the user does not agree to upload data.
        """
        if os.getenv("SUDO_USER") and os.geteuid() == 0:
            emit.progress(
                "Running with 'sudo' may cause permission errors and is discouraged.",
                permanent=True,
            )
            # Give the user a bit of time to process this before proceeding.
            time.sleep(1)

        emit.progress(
            "remote-build is experimental and is subject to change. Use with caution.",
            permanent=True,
        )

        if not parsed_args.launchpad_accept_public_upload:
            if not utils.confirm_with_user(_CONFIRMATION_PROMPT, default=False):
                emit.message("Cannot proceed without accepting a public upload.")
                return 77  # permission denied from sysexits.h

        builder = self._services.remote_build
        project = cast(models.Charm, self._services.project)
        config = cast(dict[str, Any], self.config)
        project_dir = (
            pathlib.Path(config.get("global_args", {}).get("project_dir") or ".")
            .expanduser()
            .resolve()
        )
        emit.trace(f"Project directory: {project_dir}")

        if parsed_args.launchpad_timeout:
            emit.debug(f"Setting timeout to {parsed_args.launchpad_timeout} seconds")
            builder.set_timeout(parsed_args.launchpad_timeout)

        build_id = get_build_id(self._app.name, project.name, project_dir)
        if parsed_args.recover:
            emit.progress(f"Recovering build {build_id}")
            builds = builder.resume_builds(build_id)
        else:
            emit.progress("Starting new build. It may take a while to upload large projects.")
            builds = builder.start_builds(project_dir)

        try:
            returncode = self._monitor_and_complete(build_id, builds)
        except KeyboardInterrupt:
            if utils.confirm_with_user("Cancel builds?", default=True):
                emit.progress("Cancelling builds.")
                builder.cancel_builds()
                emit.progress("Cleaning up")
                builder.cleanup()
            returncode = 0
        else:
            emit.progress("Cleaning up")
            builder.cleanup()
        return returncode

    def _monitor_and_complete(self, build_id: str | None, builds: Collection[Build]) -> int:
        builder = self._services.remote_build
        emit.progress("Monitoring build")
        try:
            for states in builder.monitor_builds():
                building: set[str] = set()
                succeeded: set[str] = set()
                uploading: set[str] = set()
                not_building: set[str] = set()
                for arch, build_state in states.items():
                    if build_state.is_running:
                        building.add(arch)
                    elif build_state == BuildState.SUCCESS:
                        succeeded.add(arch)
                    elif build_state == BuildState.UPLOADING:
                        uploading.add(arch)
                    else:
                        not_building.add(arch)
                progress_parts: list[str] = []
                if not_building:
                    progress_parts.append("Stopped: " + ",".join(sorted(not_building)))
                if building:
                    progress_parts.append("Building: " + ", ".join(sorted(building)))
                if uploading:
                    progress_parts.append("Uploading: " + ",".join(sorted(uploading)))
                if succeeded:
                    progress_parts.append("Succeeded: " + ", ".join(sorted(succeeded)))
                emit.progress("; ".join(progress_parts))
        except TimeoutError:
            if build_id:
                resume_command = f"{self._app.name} remote-build --recover --build-id={build_id}"
            else:
                resume_command = f"{self._app.name} remote-build --recover"
            emit.message(f"Timed out waiting for build.\nTo resume, run {resume_command!r}")
            return 75  # Temporary failure

        emit.progress(f"Fetching {len(builds)} build logs...")
        logs = builder.fetch_logs(pathlib.Path.cwd())

        emit.progress("Fetching build artifacts...")
        artifacts = builder.fetch_artifacts(pathlib.Path.cwd())

        log_names = sorted(path.name for path in logs.values() if path)
        artifact_names = sorted(path.name for path in artifacts)

        emit.message(
            "Build completed.\n"
            f"Log files: {', '.join(log_names)}\n"
            f"Artifacts: {', '.join(artifact_names)}"
        )
        return 0



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for resource-revisions command."""
import datetime
from argparse import Namespace
from unittest import mock

import pydantic
import pytest
from craft_store.models.resource_revision_model import (
    CharmResourceRevision,
    ResponseCharmResourceBase,
)

from charmcraft import store
from charmcraft.application.commands import ListResourceRevisionsCommand
from charmcraft.cmdbase import JSON_FORMAT
from charmcraft.env import CharmhubConfig


@pytest.fixture
def store_mock():
    """The fixture to fake the store layer in all the tests."""
    store_mock = mock.Mock(spec_set=store.Store)

    def validate_params(config, ephemeral=False, needs_auth=True):
        """Check that the store received the Charmhub configuration and ephemeral flag."""
        assert config == CharmhubConfig()
        assert isinstance(ephemeral, bool)
        assert isinstance(needs_auth, bool)
        return store_mock

    with mock.patch("charmcraft.application.commands.store.Store", validate_params):
        yield store_mock


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_resourcerevisions_simple(emitter, store_mock, config, formatted):
    """Happy path of one result from the Store."""
    store_response = [
        CharmResourceRevision(
            revision=1,
            size=pydantic.ByteSize(50),
            created_at=datetime.datetime(2020, 7, 3, 2, 30, 40, tzinfo=datetime.timezone.utc),
            bases=[ResponseCharmResourceBase()],
            name="testresource",
            sha256="",
            sha3_384="",
            sha384="",
            sha512="",
            type="file",
        ),
    ]
    store_mock.list_resource_revisions.return_value = store_response

    args = Namespace(charm_name="testcharm", resource_name="testresource", format=formatted)
    ListResourceRevisionsCommand(config).run(args)

    assert store_mock.mock_calls == [
        mock.call.list_resource_revisions("testcharm", "testresource"),
    ]
    if formatted:
        expected = [
            {
                "revision": 1,
                "created at": "2020-07-03T02:30:40+00:00",
                "size": 50,
                "bases": [{"name": "all", "channel": "all", "architectures": ["all"]}],
            },
        ]
        emitter.assert_json_output(expected)
    else:
        expected = [
            "Revision    Created at              Size  Architectures",
            "1           2020-07-03T02:30:40Z     50B  all",
        ]
        emitter.assert_messages(expected)


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_resourcerevisions_empty(emitter, store_mock, config, formatted):
    """No results from the store."""
    store_response = []
    store_mock.list_resource_revisions.return_value = store_response

    args = Namespace(charm_name="testcharm", resource_name="testresource", format=formatted)
    ListResourceRevisionsCommand(config).run(args)

    if formatted:
        emitter.assert_json_output([])
    else:
        emitter.assert_message("No revisions found.")


@pytest.mark.parametrize("formatted", [None, JSON_FORMAT])
def test_resourcerevisions_ordered_by_revision(emitter, store_mock, config, formatted):
    """Results are presented ordered by revision in the table."""
    # three Revisions with all values weirdly similar, the only difference is revision, so
    # we really assert later that it was used for ordering
    tstamp = datetime.datetime(2020, 7, 3, 20, 30, 40, tzinfo=datetime.timezone.utc)
    store_response = [
        CharmResourceRevision(
            revision=1,
            size=pydantic.ByteSize(5000),
            created_at=tstamp,
            bases=[],
            name="testresource",
            sha256="",
            sha3_384="",
            sha384="",
            sha512="",
            type="file",
        ),
        CharmResourceRevision(
            revision=3,
            size=pydantic.ByteSize(34450520),
            created_at=tstamp,
            bases=[],
            name="testresource",
            sha256="",
            sha3_384="",
            sha384="",
            sha512="",
            type="file",
        ),
        CharmResourceRevision(
            revision=4,
            size=pydantic.ByteSize(876543),
            created_at=tstamp,
            bases=[ResponseCharmResourceBase(architectures=["amd64", "arm64"])],
            name="testresource",
            sha256="",
            sha3_384="",
            sha384="",
            sha512="",
            type="file",
        ),
        CharmResourceRevision(
            revision=2,
            size=pydantic.ByteSize(50),
            created_at=tstamp,
            bases=[],
            name="testresource",
            sha256="",
            sha3_384="",
            sha384="",
            sha512="",
            type="file",
        ),
    ]
    store_mock.list_resource_revisions.return_value = store_response

    args = Namespace(charm_name="testcharm", resource_name="testresource", format=formatted)
    ListResourceRevisionsCommand(config).run(args)

    if formatted:
        expected = [
            {
                "revision": 1,
                "created at": "2020-07-03T20:30:40+00:00",
                "size": 5000,
                "bases": [],
            },
            {
                "revision": 3,
                "created at": "2020-07-03T20:30:40+00:00",
                "size": 34450520,
                "bases": [],
            },
            {
                "revision": 4,
                "created at": "2020-07-03T20:30:40+00:00",
                "size": 876543,
                "bases": [{"name": "all", "channel": "all", "architectures": ["amd64", "arm64"]}],
            },
            {"revision": 2, "created at": "2020-07-03T20:30:40+00:00", "size": 50, "bases": []},
        ]
        emitter.assert_json_output(expected)
    else:
        expected = [
            "Revision    Created at              Size  Architectures",
            "4           2020-07-03T20:30:40Z  856.0K  amd64,arm64",
            "3           2020-07-03T20:30:40Z   32.9M",
            "2           2020-07-03T20:30:40Z     50B",
            "1           2020-07-03T20:30:40Z    4.9K",
        ]
        emitter.assert_messages(expected)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Extension registry."""
from typing import Any

from charmcraft import errors
from charmcraft.extensions.extension import Extension

_EXTENSIONS: dict[str, type[Extension]] = {}


def get_extension_names() -> list[str]:
    """Obtain a extension class given the name.

    :param name: The extension name.
    :return: The list of available extensions.
    :raises ExtensionError: If the extension name is invalid.
    """
    return list(_EXTENSIONS.keys())


def get_extension_class(extension_name: str) -> type[Extension]:
    """Obtain a extension class given the name.

    :param extension_name: The extension name.
    :return: The extension class.
    :raises ExtensionError: If the extension name is invalid.
    """
    try:
        return _EXTENSIONS[extension_name]
    except KeyError:
        raise errors.ExtensionError(
            f"Extension {extension_name!r} does not exist",
            details=f"Registered extensions: {get_extension_names()}",
        ) from None


def get_extensions() -> list[dict[str, Any]]:
    """Get metadata about registered extensions."""
    return sorted(
        (
            {
                "name": name,
                "bases": [
                    "@".join(base)
                    for base in cls.get_supported_bases()
                    if not cls.is_experimental(base)
                ],
                "experimental_bases": [
                    "@".join(base)
                    for base in cls.get_supported_bases()
                    if cls.is_experimental(base)
                ],
            }
            for name, cls in _EXTENSIONS.items()
        ),
        key=lambda d: d["name"],
    )


def register(extension_name: str, extension_class: type[Extension]) -> None:
    """Register extension.

    :param extension_name: the name to register.
    :param extension_class: the Extension implementation.
    """
    _EXTENSIONS[extension_name] = extension_class


def unregister(extension_name: str) -> None:
    """Unregister extension_name.

    :raises KeyError: if extension_name is not registered.
    """
    del _EXTENSIONS[extension_name]



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the provider service."""

import pathlib

import pytest
from craft_providers import bases

from charmcraft import models, services


@pytest.fixture
def provider_service(
    fake_path: pathlib.Path,
    service_factory: services.CharmcraftServiceFactory,
    default_build_plan: list[models.CharmBuildInfo],
) -> services.ProviderService:
    fake_cache_dir = fake_path / "cache"
    fake_cache_dir.mkdir(parents=True)

    service_factory.set_kwargs(
        "provider",
        work_dir=fake_path,
        build_plan=default_build_plan,
        provider_name="host",
    )

    return service_factory.provider


@pytest.mark.parametrize(
    "base_name",
    [
        bases.BaseName("ubuntu", "20.04"),
        bases.BaseName("ubuntu", "22.04"),
        bases.BaseName("ubuntu", "24.04"),
        bases.BaseName("ubuntu", "devel"),
        pytest.param(
            bases.BaseName("centos", "7"),
            marks=[
                pytest.mark.xfail(
                    raises=AssertionError,
                    strict=True,
                    reason="https://github.com/canonical/craft-providers/issues/608",
                )
            ],
        ),
        bases.BaseName("almalinux", "9"),
    ],
)
def test_get_base_forwards_cache(
    monkeypatch,
    provider_service: services.ProviderService,
    fake_path: pathlib.Path,
    base_name: bases.BaseName,
):
    monkeypatch.setattr("charmcraft.env.get_host_shared_cache_path", lambda: fake_path / "cache")

    base = provider_service.get_base(
        base_name=base_name,
        instance_name="charmcraft-test-instance",
    )

    assert base._cache_path == fake_path / "cache"



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the store service."""


import pytest

from charmcraft import models, services


@pytest.fixture
def store_service(service_factory):
    return service_factory.store


@pytest.mark.parametrize(
    "libraries",
    [
        [models.CharmLib(lib="observability-libs.cert_handler", version="1")],
        [models.CharmLib(lib="observability-libs.cert_handler", version="1.8")],
    ],
)
def test_get_libraries(store_service: services.StoreService, libraries):
    libraries_response = store_service.get_libraries_metadata(libraries)
    assert len(libraries_response) == len(libraries)
    for lib in libraries_response:
        full_lib = store_service.get_library(
            lib.charm_name, library_id=lib.lib_id, api=lib.api, patch=lib.patch
        )
        assert full_lib.content is not None
        assert full_lib.content_hash == lib.content_hash



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for lifecycle commands."""
import argparse
import pathlib

import craft_cli
import pytest

from charmcraft import application, services
from charmcraft.application.commands import lifecycle


def get_namespace(
    *,
    bases_index=None,
    debug=False,
    destructive_mode=False,
    force=None,
    shell=False,
    shell_after=False,
    format=None,
    measure=None,
    include_all_charms: bool = False,
    include_charm: list[pathlib.Path] | None = None,
    output_bundle: pathlib.Path | None = None,
    parts: list[str] | None = None,
) -> argparse.Namespace:
    if bases_index is None:
        bases_index = []

    return argparse.Namespace(
        bases_index=bases_index,
        debug=debug,
        destructive_mode=destructive_mode,
        force=force,
        shell=shell,
        shell_after=shell_after,
        format=format,
        measure=measure,
        include_all_charms=include_all_charms,
        include_charm=include_charm,
        output_bundle=output_bundle,
        parts=parts,
    )


@pytest.fixture
def pack(service_factory: services.ServiceFactory) -> lifecycle.PackCommand:
    return lifecycle.PackCommand({"app": application.APP_METADATA, "services": service_factory})


@pytest.mark.parametrize(
    ("platform", "expected"),
    [
        ("linux", False),
        ("macos", True),
        ("win32", True),
    ],
)
def test_pack_run_managed_bundle_by_os(monkeypatch, new_path, platform, expected):
    """When packing a bundle, run_managed should return False if and only if we're on posix."""
    monkeypatch.setattr("sys.platform", platform)
    (new_path / "charmcraft.yaml").write_text("type: bundle")

    pack = lifecycle.PackCommand(None)

    result = pack.run_managed(argparse.Namespace(destructive_mode=False))

    assert result == expected


@pytest.mark.parametrize(
    ("command_args", "message_start", "project_type"),
    [
        pytest.param(
            get_namespace(include_all_charms=True),
            "--include-all-charms can only be used when packing a bundle. Currently trying ",
            "charm",
            id="include_all_charms_on_charm",
        ),
        pytest.param(
            get_namespace(include_charm=[pathlib.Path("a")]),
            "--include-charm can only be used when packing a bundle. Currently trying to pack: ",
            "charm",
            id="include_charm_on_charm",
        ),
        pytest.param(
            get_namespace(output_bundle=pathlib.Path("output.yaml")),
            "--output-bundle can only be used when packing a bundle. Currently trying to pack: ",
            "charm",
            id="output_bundle_on_charm",
        ),
    ],
)
def test_pack_invalid_arguments(
    monkeypatch,
    pack: lifecycle.PackCommand,
    command_args: argparse.Namespace,
    message_start: str,
    project_type,
) -> None:
    monkeypatch.setattr("craft_parts.utils.os_utils.OsRelease.id", lambda: "ubuntu")

    with pytest.raises(craft_cli.ArgumentParsingError) as exc_info:
        pack.run(command_args)

    assert exc_info.value.args[0].startswith(message_start)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
from textwrap import dedent
from unittest.mock import patch

import distro
import pytest
from craft_application import errors
from hypothesis import given, strategies

from charmcraft import const
from charmcraft.utils.platform import (
    OSPlatform,
    get_os_platform,
    validate_architectures,
)


@pytest.mark.parametrize(
    ("os_release", "expected_system", "expected_release"),
    [
        pytest.param(
            dedent(
                """
                # the following is an empty line

                NAME="Ubuntu"
                VERSION="20.04.1 LTS (Focal Fossa)"
                ID=ubuntu
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 20.04.1 LTS"
                VERSION_ID="20.04"
                HOME_URL="https://www.ubuntu.com/"
                SUPPORT_URL="https://help.ubuntu.com/"
                BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"

                # more in the middle; the following even would be "out of standard", but
                # we should not crash, just ignore it
                SOMETHING-WEIRD

                PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
                VERSION_CODENAME=focal
                UBUNTU_CODENAME=focal
                """
            ),
            "ubuntu",
            "20.04",
            id="ubuntu-focal",
        ),
        pytest.param(
            dedent(
                """
            NAME="Ubuntu Core"
            VERSION="20"
            ID=ubuntu-core
            PRETTY_NAME="Ubuntu Core 20"
            VERSION_ID="20"
            HOME_URL="https://snapcraft.io/"
            BUG_REPORT_URL="https://bugs.launchpad.net/snappy/"
            """
            ),
            "ubuntu-core",
            "20",
            id="ubuntu-core-20",
        ),
        pytest.param(
            dedent(
                """
                PRETTY_NAME="KDE neon 5.27"
                NAME="KDE neon"
                VERSION_ID="22.04"
                VERSION="5.27"
                VERSION_CODENAME=jammy
                ID=neon
                ID_LIKE="ubuntu debian"
                HOME_URL="https://neon.kde.org/"
                SUPPORT_URL="https://neon.kde.org/"
                BUG_REPORT_URL="https://bugs.kde.org/"
                PRIVACY_POLICY_URL="https://kde.org/privacypolicy/"
                UBUNTU_CODENAME=jammy
                LOGO=start-here-kde-neon
                """
            ),
            "ubuntu",
            "22.04",
            id="kde-neon-like-ubuntu",
        ),
    ],
)
@pytest.mark.parametrize("machine", ["x86_64", "riscv64", "arm64"])
def test_get_os_platform_linux(tmp_path, os_release, expected_system, expected_release, machine):
    """Utilize an /etc/os-release file to determine platform."""
    filepath = tmp_path / "os-release"
    filepath.write_text(os_release)
    with patch("distro.distro._distro", distro.LinuxDistribution(os_release_file=filepath)):
        with patch("platform.machine", return_value=machine):
            with patch("platform.system", return_value="Linux"):
                os_platform = get_os_platform(filepath)
    assert os_platform.system == expected_system
    assert os_platform.release == expected_release
    assert os_platform.machine == machine


@pytest.mark.parametrize("system", ["Windows", "Darwin", "Java", ""])
@pytest.mark.parametrize("release", ["NT", "Sparkling", "Jaguar", "0.0", ""])
@pytest.mark.parametrize("machine", ["AMD64", "x86_86", "arm64", "riscv64"])
def test_get_os_platform_non_linux(system, release, machine):
    """Get platform from a patched Windows machine."""
    with patch("platform.system", return_value=system):
        with patch("platform.release", return_value=release):
            with patch("platform.machine", return_value=machine):
                os_platform = get_os_platform()
    assert os_platform == OSPlatform(system, release, machine)


@given(strategies.iterables(strategies.sampled_from(sorted(const.SUPPORTED_ARCHITECTURES))))
def test_validate_architectures_valid_values(architectures):
    validate_architectures(architectures)


@given(strategies.iterables(strategies.just("all")))
def test_validate_architectures_all_success(architectures):
    validate_architectures(architectures, allow_all=True)


@pytest.mark.parametrize(
    ("architectures", "message"),
    [
        (["brutalist"], "Invalid architecture(s): brutalist"),
    ],
)
@pytest.mark.parametrize(
    "details",
    [
        "Valid architecture values are: 'amd64', 'arm64', 'armhf', 'ppc64el', 'riscv64', and 's390x'"
    ],
)
def test_validate_architectures_error(architectures, message, details):
    match = r"^Invalid architecture\(s\): "
    with pytest.raises(errors.CraftValidationError, match=match) as exc_info:
        validate_architectures(architectures)

    assert exc_info.value.details == details


def test_validate_architectures_all_error():
    with pytest.raises(
        errors.CraftValidationError,
        match="If 'all' is defined for architectures, it must be the only architecture.",
    ):
        validate_architectures(["all", "amd64"], allow_all=True)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import pathlib
import tempfile

import pytest

from charmcraft.errors import MissingDependenciesError
from charmcraft.utils.package import (
    exclude_packages,
    get_package_names,
    get_pip_command,
    get_pip_version,
    get_pypi_packages,
    get_requirements_file_package_names,
    validate_strict_dependencies,
)


@pytest.mark.parametrize(
    ("requirements", "expected"),
    [
        pytest.param([], set(), id="empty"),
        pytest.param(["abc==1.0.0"], {"abc==1.0.0"}, id="simple"),
        pytest.param(["-e ."], set(), id="editable-ignored"),
        pytest.param([""], set(), id="empty-line"),
    ],
)
def test_get_pypi_packages(requirements, expected):
    assert get_pypi_packages(requirements) == expected


@pytest.mark.parametrize(
    ("packages", "expected"),
    [
        # Specifiers from pep440: https://peps.python.org/pep-0440/#version-specifiers
        pytest.param({"abc"}, {"abc"}, id="no-version"),
        pytest.param({"abc==1.0.0"}, {"abc"}, id="version-matching"),
        pytest.param({"abc >= 1.0.0"}, {"abc"}, id="inclusive-ordered-gt"),
        pytest.param({"abc<= 1.0.0"}, {"abc"}, id="inclusive-ordered-lt"),
        pytest.param({"abc ~= 1.0"}, {"abc"}, id="compatible-release"),
        pytest.param({"abc===foobar"}, {"abc"}, id="arbitrary-equality"),
        pytest.param({"abc >=1.0,<2.0, !=1.2.3.*"}, {"abc"}, id="compound-specifier"),
    ],
)
def test_get_package_names(packages, expected):
    assert get_package_names(packages) == expected


@pytest.mark.parametrize(
    ("requirements", "excluded", "expected"),
    [
        pytest.param(set(), set(), set(), id="empty"),
        pytest.param({"abc==1.0.0"}, {"abc"}, set(), id="make-empty"),
        pytest.param({"abc==1.0.0", "def==1.2.3"}, {"abc"}, {"def==1.2.3"}, id="remove-one"),
        pytest.param({"abc==1.0.0"}, {"invalid"}, {"abc==1.0.0"}, id="irrelevant-exclusion"),
    ],
)
def test_exclude_packages(requirements, excluded, expected):
    assert exclude_packages(requirements, excluded=excluded) == expected


@pytest.mark.parametrize(
    ("file_contents", "expected"),
    [
        pytest.param([], set(), id="empty"),
        pytest.param(["abc>=1.3.5", "abc<2.0"], {"abc"}, id="versions_stripped"),
        pytest.param(["abc>1.0", "abc<0.1"], {"abc"}, id="versions_not_checked"),
    ],
)
def test_get_requirements_file_package_names(tmp_path, file_contents, expected):
    """Test that get_requirements_file_package_names succeeds with correct outputs."""
    files = []
    for n, content in enumerate(file_contents):
        current_file = tmp_path / f"requirements-{n}.txt"
        current_file.write_text(content)
        files.append(current_file)

    assert get_requirements_file_package_names(*files) == expected


@pytest.mark.parametrize(
    (
        "requirements",
        "source_deps",
        "binary_deps",
        "expected_no_binary",
        "expected_other_packages",
    ),
    [
        (["abc==1.0.0", "def>=1.2.3"], [], ["def"], "--no-binary=abc", []),
        (
            ["abc==1.0", "def>=1.2.3"],
            ["ghi"],
            ["def", "jkl"],
            "--no-binary=abc,ghi",
            ["ghi", "jkl"],
        ),
        (["abc==1.0.0", "def>=1.2.3"], [], [], "--no-binary=:all:", []),
        ([], ["abc==1.0.0", "def>=1.2.3"], [], "--no-binary=:all:", ["abc==1.0.0", "def>=1.2.3"]),
    ],
)
@pytest.mark.parametrize("prefix", [["/bin/pip"], ["/some/path/to/pip3"], ["pip", "--some-param"]])
def test_get_pip_command(
    prefix, requirements, source_deps, binary_deps, expected_no_binary, expected_other_packages
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = pathlib.Path(tmp_dir, "requirements.txt")
        path.write_text("\n".join(requirements))

        command = get_pip_command(prefix, [path], source_deps=source_deps, binary_deps=binary_deps)
        assert command[: len(prefix)] == prefix
        actual_no_binary, actual_requirement, *actual_other_packgaes = command[len(prefix) :]
        assert actual_no_binary == expected_no_binary
        assert actual_other_packgaes == expected_other_packages
        assert actual_requirement == f"--requirement={path}"


@pytest.mark.parametrize(
    ("pip_cmd", "stdout", "expected"),
    [("pip", "pip 22.0.2 from /usr/lib/python3/dist-packages/pip (python 3.10)\n", (22, 0, 2))],
)
def test_get_pip_version_success(
    fake_process,
    pip_cmd,
    stdout,
    expected,
):
    fake_process.register([pip_cmd, "--version"], stdout=stdout)

    assert get_pip_version(pip_cmd) == expected


@pytest.mark.parametrize(
    ("pip_cmd", "stdout", "error_msg"),
    [
        ("pip", "pip?", "Unknown pip version"),
        ("pip", "pip 1.0.0-dev0-yolo", "Unknown pip version 1.0.0-dev0-yolo"),
    ],
)
def test_get_pip_version_parsing_failure(fake_process, pip_cmd, stdout, error_msg):
    fake_process.register([pip_cmd, "--version"], stdout=stdout)

    with pytest.raises(ValueError) as exc_info:
        get_pip_version(pip_cmd)

    assert exc_info.value.args[0] == error_msg


@pytest.mark.parametrize(
    ("dependencies", "other_packages"),
    [
        ([], []),
        (["abc>2.0", "def<1.0"], ["abc", "def"]),
        (["abc", "def", "ghi"], []),
    ],
)
def test_validate_strict_dependencies_success(dependencies, other_packages):
    validate_strict_dependencies(dependencies, other_packages)


@pytest.mark.parametrize(
    ("dependencies", "other_packages", "extra_packages"),
    [
        ([], ["abc"], ["abc"]),
        (["abc>=1.0"], ["abc<2.0", "def>1.2"], ["def"]),
        ([], ["zyx", "wvut"], ["wvut", "zyx"]),
    ],
)
def test_validate_strict_dependencies_missing(dependencies, other_packages, extra_packages):
    with pytest.raises(MissingDependenciesError) as exc_info:
        validate_strict_dependencies(dependencies, other_packages)

    assert exc_info.value.extra_dependencies == extra_packages
